from __future__ import annotations

from enum import Enum


class State(Enum):
    """Indicate the state of a sequence.

    Each sequence in the experiment session has a state. Only sequences in the DRAFT
    state can have their configuration modified. Only sequences in the RUNNING state
    can have data added to them.

    Attributes:
        DRAFT: The sequence was not started yet and is still being edited.
            When the sequence is passed to the experiment manager, it will transition
            to the PREPARING state.
        PREPARING: The sequence is being prepared to run.
            It is currently acquiring the necessary resources and devices.
            No shot have been scheduled to run yet.
            The sequence cannot be interrupted while it is preparing.
            If the preparation succeeds, the sequence will transition to the RUNNING
            state.
            If the preparation fails, the sequence will transition to the CRASHED state.
        RUNNING: The sequence is currently running shots.
            When all shots are completed, the sequence will transition to the FINISHED
            state.
            If the sequence is interrupted, it will transition to the INTERRUPTED state.
            If an error occurs while running the sequence, it will transition to the
            CRASHED state.
        FINISHED: The sequence was successfully run and all shots were completed.
        INTERRUPTED: The user interrupted the sequence while it was running.
        CRASHED: An error occurred while running the sequence.

    The states FINISHED, INTERRUPTED and CRASHED are terminal states. Once a sequence is
    in one of these states, the only way to change its state is to erase all associated
    data and reset it back to the DRAFT state.
    """

    DRAFT = "draft"
    PREPARING = "preparing"
    RUNNING = "running"
    FINISHED = "finished"
    INTERRUPTED = "interrupted"
    CRASHED = "crashed"

    @classmethod
    def is_transition_allowed(cls, old_state: State, new_state: State) -> bool:
        return new_state in _ALLOWED_TRANSITIONS[old_state]

    def is_editable(self) -> bool:
        """Indicate if a sequence in this state can be edited."""

        return self in {State.DRAFT}

    def __str__(self):
        return self.value


_ALLOWED_TRANSITIONS = {
    State.DRAFT: {State.PREPARING},
    State.PREPARING: {State.RUNNING, State.CRASHED},
    State.RUNNING: {State.FINISHED, State.INTERRUPTED, State.CRASHED},
    State.FINISHED: {State.DRAFT},
    State.INTERRUPTED: {State.DRAFT},
    State.CRASHED: {State.DRAFT},
}



from typing import NewType, TypeGuard

DeviceName = NewType("DeviceName", str)
"""A non-empty string that uniquely identifies a device."""


def is_device_name(name: str) -> TypeGuard[DeviceName]:
    return isinstance(name, str) and len(name) > 0



from .analog_time_lane import AnalogTimeLane, Ramp
from .camera_time_lane import CameraTimeLane, TakePicture
from .digital_time_lane import DigitalTimeLane
from .timelane import TimeLane, TimeLanes, TimeLaneType, Block, Step, Span

__all__ = [
    "TimeLane",
    "TimeLanes",
    "DigitalTimeLane",
    "AnalogTimeLane",
    "Ramp",
    "CameraTimeLane",
    "TakePicture",
    "TimeLaneType",
    "Block",
    "Step",
    "Span",
]



#!/usr/bin/python
from PySide6 import QtWidgets, QtCore


class PropLabel(QtWidgets.QLabel):
    """
    Displays a node property as a "QLabel" widget in the PropertiesBin widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropLabel, self).__init__(parent)
        self._name = None

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.text()

    def set_value(self, value):
        if value != self.get_value():
            self.setText(str(value))
            self.value_changed.emit(self.get_name(), value)


class PropLineEdit(QtWidgets.QLineEdit):
    """
    Displays a node property as a "QLineEdit" widget in the PropertiesBin
    widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropLineEdit, self).__init__(parent)
        self._name = None
        self.editingFinished.connect(self._on_editing_finished)

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def _on_editing_finished(self):
        self.value_changed.emit(self.get_name(), self.text())

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.text()

    def set_value(self, value):
        _value = str(value)
        if _value != self.get_value():
            self.setText(_value)
            self.value_changed.emit(self.get_name(), _value)


class PropTextEdit(QtWidgets.QTextEdit):
    """
    Displays a node property as a "QTextEdit" widget in the PropertiesBin
    widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropTextEdit, self).__init__(parent)
        self._name = None
        self._prev_text = ""

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def focusInEvent(self, event):
        super(PropTextEdit, self).focusInEvent(event)
        self._prev_text = self.toPlainText()

    def focusOutEvent(self, event):
        super(PropTextEdit, self).focusOutEvent(event)
        if self._prev_text != self.toPlainText():
            self.value_changed.emit(self.get_name(), self.toPlainText())
        self._prev_text = ""

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.toPlainText()

    def set_value(self, value):
        _value = str(value)
        if _value != self.get_value():
            self.setPlainText(_value)
            self.value_changed.emit(self.get_name(), _value)


class PropComboBox(QtWidgets.QComboBox):
    """
    Displays a node property as a "QComboBox" widget in the PropertiesBin
    widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropComboBox, self).__init__(parent)
        self._name = None
        self.currentIndexChanged.connect(self._on_index_changed)

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def _on_index_changed(self):
        self.value_changed.emit(self.get_name(), self.get_value())

    def items(self):
        """
        Returns items from the combobox.

        Returns:
            list[str]: list of strings.
        """
        return [self.itemText(i) for i in range(self.count())]

    def set_items(self, items):
        """
        Set items on the combobox.

        Args:
            items (list[str]): list of strings.
        """
        self.clear()
        self.addItems(items)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.currentText()

    def set_value(self, value):
        if value != self.get_value():
            idx = self.findText(value, QtCore.Qt.MatchExactly)
            self.setCurrentIndex(idx)
            if idx >= 0:
                self.value_changed.emit(self.get_name(), value)


class PropCheckBox(QtWidgets.QCheckBox):
    """
    Displays a node property as a "QCheckBox" widget in the PropertiesBin
    widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropCheckBox, self).__init__(parent)
        self._name = None
        self.clicked.connect(self._on_clicked)

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def _on_clicked(self):
        self.value_changed.emit(self.get_name(), self.get_value())

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.isChecked()

    def set_value(self, value):
        _value = bool(value)
        if _value != self.get_value():
            self.setChecked(_value)
            self.value_changed.emit(self.get_name(), _value)


class PropSpinBox(QtWidgets.QSpinBox):
    """
    Displays a node property as a "QSpinBox" widget in the PropertiesBin widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropSpinBox, self).__init__(parent)
        self._name = None
        self.setButtonSymbols(self.NoButtons)
        self.valueChanged.connect(self._on_value_change)

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def _on_value_change(self, value):
        self.value_changed.emit(self.get_name(), value)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.value()

    def set_value(self, value):
        if value != self.get_value():
            self.setValue(value)


class PropDoubleSpinBox(QtWidgets.QDoubleSpinBox):
    """
    Displays a node property as a "QDoubleSpinBox" widget in the PropertiesBin
    widget.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropDoubleSpinBox, self).__init__(parent)
        self._name = None
        self.setButtonSymbols(self.NoButtons)
        self.valueChanged.connect(self._on_value_change)

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def _on_value_change(self, value):
        self.value_changed.emit(self.get_name(), value)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def get_value(self):
        return self.value()

    def set_value(self, value):
        if value != self.get_value():
            self.setValue(value)


# class PropPushButton(QtWidgets.QPushButton):
#     """
#     Displays a node property as a "QPushButton" widget in the PropertiesBin
#     widget.
#     """
#
#     value_changed = QtCore.Signal(str, object)
#     button_clicked = QtCore.Signal(str, object)
#
#     def __init__(self, parent=None):
#         super(PropPushButton, self).__init__(parent)
#         self._name = None
#         self.clicked.connect(self.button_clicked.emit)
#
#     def set_on_click_func(self, func, node):
#         """
#         Sets slot function for the PropPushButton widget.
#
#         Args:
#             func (function): property slot function.
#             node (NodeGraphQt.NodeObject): node object.
#         """
#         if not callable(func):
#             raise TypeError('var func is not a function.')
#         self.clicked.connect(lambda: func(node))
#
#     def get_value(self):
#         return
#
#     def set_value(self, value):
#         return



import contextlib

import attrs
import sqlalchemy.orm

from caqtus.types.parameter import ParameterNamespace
from caqtus.utils import serialization
from ._device_configuration_collection import SQLDeviceConfigurationCollection
from ._path_hierarchy import SQLPathHierarchy
from ._sequence_collection import (
    SQLSequenceCollection,
)
from ._serializer import SerializerProtocol
from .parameters_table import SQLParameters
from .._experiment_session import (
    ExperimentSession,
    ExperimentSessionNotActiveError,
)


@attrs.frozen
class Active:
    session: sqlalchemy.orm.Session
    session_context: contextlib.AbstractContextManager[sqlalchemy.orm.Session]

    def __str__(self):
        return f"active"


@attrs.frozen
class Inactive:
    session_context: contextlib.AbstractContextManager[sqlalchemy.orm.Session]

    def __str__(self):
        return f"inactive"


@attrs.frozen
class Closed:
    def __str__(self):
        return f"closed"


SessionState = Inactive | Active | Closed


class SQLExperimentSession(ExperimentSession):
    """Used to store experiment data in a SQL database.

    This class implements the :class:`ExperimentSession` interface and the documentation
    of the related methods can be found in the :class:`ExperimentSession` documentation.
    """

    def __init__(
        self,
        session_context: contextlib.AbstractContextManager[sqlalchemy.orm.Session],
        serializer: SerializerProtocol,
        *args,
        **kwargs,
    ):
        """Create a new experiment session.

        This constructor is not meant to be called directly.
        Instead, use a :py:class:`SQLExperimentSessionMaker` to create a new session.
        """

        super().__init__(*args, **kwargs)
        self._state = Inactive(session_context=session_context)
        self.paths = SQLPathHierarchy(parent_session=self)
        self.sequences = SQLSequenceCollection(
            parent_session=self, serializer=serializer
        )
        self.default_device_configurations = SQLDeviceConfigurationCollection(
            parent_session=self, serializer=serializer
        )

    def get_global_parameters(self) -> ParameterNamespace:
        return _get_global_parameters(self._get_sql_session())

    def set_global_parameters(self, parameters: ParameterNamespace) -> None:
        return _set_global_parameters(self._get_sql_session(), parameters)

    def __repr__(self):
        return f"<{self.__class__.__name__} ({self._state})>"

    def __enter__(self):
        if not isinstance(self._state, Inactive):
            error = RuntimeError("Session has already been activated")
            error.add_note(
                "You cannot reactivate a session, you must create a new one."
            )
            raise error
        context = self._state.session_context
        sql_session = context.__enter__()
        self._state = Active(session=sql_session, session_context=context)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not isinstance(self._state, Active):
            raise ExperimentSessionNotActiveError(
                "Experiment session is not active and cannot be exited"
            )
        context = self._state.session_context
        context.__exit__(exc_type, exc_val, exc_tb)
        self._state = Closed()

    def _get_sql_session(self) -> sqlalchemy.orm.Session:
        if not isinstance(self._state, Active):
            raise ExperimentSessionNotActiveError(
                "Experiment session is not active and cannot be used"
            )
        return self._state.session


def _get_global_parameters(session: sqlalchemy.orm.Session) -> ParameterNamespace:
    stmt = sqlalchemy.select(SQLParameters).where(SQLParameters.name == "global")
    result = session.execute(stmt)
    if found := result.scalar():
        return serialization.converters["json"].structure(
            found.content, ParameterNamespace
        )
    else:
        # It could be that the table is empty if set_global_parameters was never
        # called before, in which case we return an empty ParameterNamespace.
        return ParameterNamespace.empty()


def _set_global_parameters(
    session: sqlalchemy.orm.Session, parameters: ParameterNamespace
) -> None:
    if not isinstance(parameters, ParameterNamespace):
        raise TypeError(
            f"Expected a ParameterNamespace, got {type(parameters).__name__}"
        )
    query = sqlalchemy.select(SQLParameters).where(SQLParameters.name == "global")
    result = session.execute(query)
    content = serialization.converters["json"].unstructure(
        parameters, ParameterNamespace
    )
    if found := result.scalar():
        found.content = content
    else:
        new_parameters = SQLParameters(name="global", content=content)
        session.add(new_parameters)



from __future__ import annotations

import abc
import bisect
import collections
import itertools
import math
from collections.abc import Sequence
from typing import (
    NewType,
    TypeVar,
    overload,
    Optional,
    assert_never,
    Callable,
)

import numpy
import numpy as np
import numpy.typing as npt
from numpy.typing import DTypeLike

Length = NewType("Length", int)
Width = NewType("Width", int)
Depth = NewType("Depth", int)

_S = TypeVar("_S", covariant=True, bound=DTypeLike)

type Array1D[T: np.generic] = npt.NDArray[T]


class SequencerInstruction[T: np.generic](abc.ABC):
    """An immutable representation of instructions to output on a sequencer.

    This represents a high-level series of instructions to output on a sequencer.
    Each instruction is a compact representation of values to output at integer time
    steps.
    The length of the instruction is the number of time steps it takes to output all
    the values.
    The width of the instruction is the number of channels that are output at each time
    step.

    Instructions can be concatenated in time using the `+` operator or the
    :func:`concatenate`.
    An instruction can be repeated using the `*` operator with an integer.
    """

    @abc.abstractmethod
    def __len__(self) -> Length:
        """Returns the length of the instruction in clock cycles."""

        raise NotImplementedError

    @overload
    @abc.abstractmethod
    def __getitem__(self, item: int) -> T:
        """Returns the value at the given index."""

        ...

    @overload
    @abc.abstractmethod
    def __getitem__(self, item: slice) -> SequencerInstruction[T]:
        """Returns a sub-instruction over the given slice.

        Warning:
            Not all valid slices are supported.
            Only slices with a step of 1 are fully supported for all instructions.
        """

        ...

    @overload
    @abc.abstractmethod
    def __getitem__(self, item: str) -> SequencerInstruction:
        """Returns a sub-instruction over the given field.

        Returns:
            A new instruction with the given field.
            This new instruction has the same length as the original instruction.

        Raises:
            ValueError: If the instruction does not have fields or the field is not
            found.
        """

        ...

    @abc.abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dtype(self) -> numpy.dtype[T]:
        """Returns the dtype of the instruction."""

        raise NotImplementedError

    @abc.abstractmethod
    def as_type[S: np.generic](self, dtype: numpy.dtype[S]) -> SequencerInstruction[S]:
        """Returns a new instruction with the given dtype."""

        raise NotImplementedError

    @property
    def width(self) -> Width:
        """Returns the number of parallel channels that are output at each time step."""

        fields = self.dtype.fields
        if fields is None:
            return Width(1)
        else:
            return Width(len(fields))

    @property
    @abc.abstractmethod
    def depth(self) -> Depth:
        """Returns the number of nested instructions.

        The invariant `instruction.depth <= len(instruction)` always holds.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def to_pattern(self) -> Pattern[T]:
        """Returns a flattened pattern of the instruction."""

        raise NotImplementedError

    @abc.abstractmethod
    def __eq__(self, other):
        raise NotImplementedError

    def __add__(self, other) -> SequencerInstruction[T]:
        if isinstance(other, SequencerInstruction):
            if len(self) == 0:
                return other
            elif len(other) == 0:
                return self
            else:
                return concatenate(self, other)
        else:
            return NotImplemented

    def __mul__(self, other) -> SequencerInstruction[T]:
        if isinstance(other, int):
            if other < 0:
                raise ValueError("Repetitions must be a positive integer")
            elif other == 0:
                return empty_like(self)
            elif other == 1:
                return self
            else:
                if isinstance(self, Repeated):
                    return Repeated(self._repetitions * other, self._instruction)
                else:
                    return Repeated(other, self)
        else:
            # We specifically raise an error here and not return NotImplemented to avoid
            # multiplication by a numpy integer taking over and returning a numpy
            # array instead of a SequencerInstruction.
            raise TypeError(f"Cannot multiply instruction by {other!r}")

    def __rmul__(self, other) -> SequencerInstruction[T]:
        return self.__mul__(other)

    @abc.abstractmethod
    def apply[
        S: np.generic
    ](self, func: Callable[[Array1D[T]], Array1D[S]]) -> SequencerInstruction[S]:
        """Applies an element-wise function to the values of the instruction.

        Args:
            func: The function to apply to the values of the instruction.
                It must take a 1D numpy array with the same dtype as the instruction
                and return a 1D numpy array with the same length and a (possibly)
                different dtype.

        Returns:
            A new instruction with the function applied to the values.
            The length of the new instruction is the same as the original instruction.

        Raises:
            ValueError: If the function does not return an array of the same length as
                the one given.
        """

        raise NotImplementedError

    def _repr_mimebundle_(self, include=None, exclude=None):
        from ._to_graph import to_graph

        graph = to_graph(self)
        return graph._repr_mimebundle_(include, exclude)


class Pattern[T: np.generic](SequencerInstruction[T]):
    """An instruction representing a sequence of values.

    This is a fully explicit instruction for which each sample point must be given.

    Args:
        pattern: The sequence of values that this pattern represents.
        dtype: The dtype of the pattern.
            If not provided, it is inferred from the values.

    Raises:
        ValueError: If the pattern contains non-finite values.
    """

    # All values inside the pattern MUST be finite (no NaN, no inf).
    # This is ensured by public methods, but not necessarily by all private methods.

    __slots__ = ("_pattern", "_length")

    def __init__(self, pattern: npt.ArrayLike, dtype: Optional[np.dtype[T]] = None):
        self._pattern = numpy.array(pattern, dtype=dtype)
        if not _has_only_finite_values(self._pattern):
            raise ValueError("Pattern must contain only finite values")
        self._pattern.setflags(write=False)
        self._length = Length(len(self._pattern))

    def __repr__(self):
        if np.issubdtype(self.dtype, np.void):
            return f"Pattern({self._pattern.tolist()!r}, dtype={self.dtype})"
        else:
            return f"Pattern({self._pattern.tolist()!r})"

    def __str__(self):
        return str(self._pattern.tolist())

    @overload
    def __getitem__(self, item: int) -> T: ...

    @overload
    def __getitem__(self, item: slice) -> Pattern[T]: ...

    @overload
    def __getitem__(self, item: str) -> SequencerInstruction: ...

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._pattern[item]
        elif isinstance(item, slice):
            return Pattern.create_without_copy(self._pattern[item])
        elif isinstance(item, str):
            return Pattern.create_without_copy(self._pattern[item])
        else:
            assert_never(item)

    @classmethod
    def create_without_copy[S: np.generic](cls, array: Array1D[S]) -> Pattern[S]:
        if not _has_only_finite_values(array):
            raise ValueError("Pattern must contain only finite values")
        array.setflags(write=False)
        pattern = cls.__new__(cls)
        pattern._pattern = array
        pattern._length = Length(len(array))
        return pattern  # type: ignore

    @property
    def dtype(self) -> numpy.dtype[T]:
        return self._pattern.dtype

    def as_type[S: np.generic](self, dtype: numpy.dtype[S]) -> Pattern[S]:
        return Pattern.create_without_copy(self._pattern.astype(dtype, copy=False))

    def __len__(self) -> Length:
        return self._length

    @property
    def depth(self) -> Depth:
        return Depth(0)

    def to_pattern(self) -> Pattern[T]:
        return self

    def __eq__(self, other):
        if isinstance(other, Pattern):
            return numpy.array_equal(self._pattern, other._pattern)
        else:
            return NotImplemented

    def apply[
        S: np.generic
    ](self, func: Callable[[Array1D[T]], Array1D[S]]) -> Pattern[S]:
        result = func(self._pattern)
        if len(result) != len(self):
            raise ValueError("Function must return an array of the same length")
        if not _has_only_finite_values(result):
            raise ValueError("Function must return an array with only finite values")
        return Pattern.create_without_copy(result)

    @property
    def array(self) -> Array1D[T]:
        return self._pattern


def _has_only_finite_values[T: np.generic](array: Array1D[T]) -> bool:
    if np.issubdtype(array.dtype, np.floating):
        return bool(np.all(np.isfinite(array)))
    else:
        return True


class Concatenated[T: np.generic](SequencerInstruction[T]):
    """Represents an immutable concatenation of instructions.

    Use the `+` operator or the function :func:`concatenate` to concatenate instructions.
    Do not use the class constructor directly.
    """

    __slots__ = ("_instructions", "_instruction_bounds", "_length")
    __match_args__ = ("instructions",)

    @property
    def instructions(self) -> tuple[SequencerInstruction[T], ...]:
        """The instructions concatenated by this instruction."""

        return self._instructions

    def __init__(self, *instructions: SequencerInstruction[T]):
        assert all(
            isinstance(instruction, SequencerInstruction)
            for instruction in instructions
        )
        # The following assertions define a "pure" concatenation.
        # (i.e. no empty instructions, no nested concatenations, and at least two
        # instructions).
        assert all(len(instruction) >= 1 for instruction in instructions)
        assert len(instructions) >= 2
        assert all(
            not isinstance(instruction, Concatenated) for instruction in instructions
        )

        assert all(
            instruction.dtype == instructions[0].dtype for instruction in instructions
        )

        self._instructions = instructions

        # self._instruction_bounds[i] is the first element index (included) the i-th
        # instruction
        #
        # self._instruction_bounds[i+1] is the last element index (excluded) of the
        # i-th instruction
        self._instruction_bounds = (0,) + tuple(
            itertools.accumulate(len(instruction) for instruction in self._instructions)
        )
        self._length = Length(self._instruction_bounds[-1])

    def __repr__(self):
        inner = ", ".join(repr(instruction) for instruction in self._instructions)
        return f"Concatenated({inner})"

    def __str__(self):
        sub_strings = [str(instruction) for instruction in self._instructions]
        return " + ".join(sub_strings)

    @overload
    def __getitem__(self, item: int) -> T: ...

    @overload
    def __getitem__(self, item: slice) -> SequencerInstruction[T]: ...

    @overload
    def __getitem__(self, item: str) -> SequencerInstruction: ...

    def __getitem__(self, item):
        match item:
            case int() as index:
                return self._get_index(index)
            case slice() as slice_:
                return self._get_slice(slice_)
            case str() as field:
                return self._get_field(field)
            case _:
                assert_never(item)

    def _get_index(self, index: int) -> T:
        index = _normalize_index(index, len(self))
        instruction_index = bisect.bisect_right(self._instruction_bounds, index) - 1
        instruction = self._instructions[instruction_index]
        instruction_start_index = self._instruction_bounds[instruction_index]
        return instruction[index - instruction_start_index]

    def _get_slice(self, slice_: slice) -> SequencerInstruction[T]:
        start, stop, step = _normalize_slice(slice_, len(self))
        if step != 1:
            raise NotImplementedError
        start_step_index = bisect.bisect_right(self._instruction_bounds, start) - 1
        stop_step_index = bisect.bisect_left(self._instruction_bounds, stop) - 1

        results: list[SequencerInstruction[T]] = [empty_like(self)]
        for instruction_index in range(start_step_index, stop_step_index + 1):
            instruction_start_index = self._instruction_bounds[instruction_index]
            instruction_slice_start = max(start, instruction_start_index)
            instruction_stop_index = self._instruction_bounds[instruction_index + 1]
            instruction_slice_stop = min(stop, instruction_stop_index)
            instruction_slice = slice(
                instruction_slice_start - instruction_start_index,
                instruction_slice_stop - instruction_start_index,
                step,
            )
            results.append(self._instructions[instruction_index][instruction_slice])
        return concatenate(*results)

    def _get_field(self, field: str) -> SequencerInstruction:
        return Concatenated(*(instruction[field] for instruction in self._instructions))

    @property
    def dtype(self) -> numpy.dtype[T]:
        return self._instructions[0].dtype

    def as_type[S: np.generic](self, dtype: numpy.dtype[S]) -> Concatenated[S]:
        return Concatenated[S](
            *(instruction.as_type(dtype) for instruction in self._instructions)
        )

    def __len__(self) -> Length:
        return self._length

    @property
    def depth(self) -> Depth:
        return Depth(max(instruction.depth for instruction in self._instructions) + 1)

    def to_pattern(self) -> Pattern[T]:
        # noinspection PyProtectedMember
        new_array = numpy.concatenate(
            [instruction.to_pattern()._pattern for instruction in self._instructions],
            casting="safe",
        )
        return Pattern.create_without_copy(new_array)

    def __eq__(self, other):
        if isinstance(other, Concatenated):
            return self._instructions == other._instructions
        else:
            return NotImplemented

    def apply[
        S: np.generic
    ](self, func: Callable[[Array1D[T]], Array1D[S]]) -> Concatenated[S]:
        return Concatenated(
            *(instruction.apply(func) for instruction in self._instructions)
        )


class Repeated[T: np.generic](SequencerInstruction[T]):
    """Represents a repetition of an instruction.

    Use the `*` operator with an integer to repeat an instruction.
    Do not use the class constructor directly.

    Attributes:
        instruction: The instruction to repeat.
        repetitions: The number of times to repeat the instruction.
    """

    __slots__ = ("_repetitions", "_instruction", "_length")

    @property
    def repetitions(self) -> int:
        return self._repetitions

    @property
    def instruction(self) -> SequencerInstruction[T]:
        return self._instruction

    def __init__(self, repetitions: int, instruction: SequencerInstruction[T]):
        """
        Do not use this constructor in user code.
        Instead, use the `*` operator.
        """

        assert isinstance(repetitions, int)
        assert isinstance(instruction, SequencerInstruction)
        assert repetitions >= 2
        assert len(instruction) >= 1
        assert not isinstance(instruction, Repeated)

        self._repetitions = repetitions
        self._instruction = instruction
        self._length = Length(len(self._instruction) * self._repetitions)

    def __repr__(self):
        return (
            f"Repeated(repetitions={self._repetitions!r},"
            f" instruction={self._instruction!r})"
        )

    def __str__(self):
        if isinstance(self._instruction, Concatenated):
            return f"{self._repetitions} * ({self._instruction!s})"
        else:
            return f"{self._repetitions} * {self._instruction!s}"

    def __len__(self) -> Length:
        return self._length

    @overload
    def __getitem__(self, item: int) -> T: ...

    @overload
    def __getitem__(self, item: slice) -> SequencerInstruction[T]: ...

    @overload
    def __getitem__(self, item: str) -> SequencerInstruction: ...

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_index(item)
        elif isinstance(item, slice):
            return self._get_slice(item)
        elif isinstance(item, str):
            return self._get_field(item)
        else:
            assert_never(item)

    def _get_index(self, index: int) -> T:
        index = _normalize_index(index, len(self))
        _, r = divmod(index, len(self._instruction))
        return self._instruction[r]

    def _get_slice(self, slice_: slice) -> SequencerInstruction[T]:
        start, stop, step = _normalize_slice(slice_, len(self))
        if step != 1:
            raise NotImplementedError
        length = len(self._instruction)
        first_repetition = math.ceil(start / length)
        last_repetition = math.floor(stop / length)
        if first_repetition > last_repetition:
            return self._instruction[
                start - first_repetition * length : stop - first_repetition * length
            ]
        else:
            previous_repetition = math.floor(start / length)
            prepend = self._instruction[
                start
                - previous_repetition
                * length : (first_repetition - previous_repetition)
                * length
            ]
            middle = self._instruction * (last_repetition - first_repetition)
            append = self._instruction[: stop - last_repetition * length]
            return prepend + middle + append

    def _get_field(self, field: str) -> SequencerInstruction:
        return Repeated(self._repetitions, self._instruction[field])

    @property
    def dtype(self) -> numpy.dtype[T]:
        return self._instruction.dtype

    def as_type[S: np.generic](self, dtype: numpy.dtype[S]) -> Repeated[S]:
        return Repeated(self._repetitions, self._instruction.as_type(dtype))

    @property
    def depth(self) -> Depth:
        return Depth(self._instruction.depth + 1)

    def to_pattern(self) -> Pattern[T]:
        inner_pattern = self._instruction.to_pattern()
        # noinspection PyProtectedMember
        new_array = numpy.tile(inner_pattern._pattern, self._repetitions)
        return Pattern.create_without_copy(new_array)

    def __eq__(self, other):
        if isinstance(other, Repeated):
            return (
                self._repetitions == other._repetitions
                and self._instruction == other._instruction
            )
        else:
            return NotImplemented

    def apply[
        S: np.generic
    ](self, func: Callable[[Array1D[T]], Array1D[S]]) -> Repeated[S]:
        return Repeated(self._repetitions, self._instruction.apply(func))


def _normalize_index(index: int, length: int) -> int:
    normalized = index if index >= 0 else length + index
    if not 0 <= normalized < length:
        raise IndexError(f"Index {index} is out of bounds for length {length}")
    return normalized


def _normalize_slice_index(index: int, length: int) -> int:
    normalized = index if index >= 0 else length + index
    if not 0 <= normalized <= length:
        raise IndexError(f"Slice index {index} is out of bounds for length {length}")
    return normalized


def _normalize_slice(slice_: slice, length: int) -> tuple[int, int, int]:
    step = slice_.step or 1
    if step == 0:
        raise ValueError("Slice step cannot be zero")
    if slice_.start is None:
        start = 0 if step > 0 else length - 1
    else:
        start = _normalize_slice_index(slice_.start, length)
    if slice_.stop is None:
        stop = length if step > 0 else -1
    else:
        stop = _normalize_slice_index(slice_.stop, length)

    return start, stop, step


def empty_like[T: np.generic](instruction: SequencerInstruction[T]) -> Pattern[T]:
    return empty_with_dtype(instruction.dtype)


def empty_with_dtype[T: np.generic](dtype: numpy.dtype[T]) -> Pattern[T]:
    return Pattern([], dtype=dtype)


def concatenate[
    T: np.generic
](*instructions: SequencerInstruction[T]) -> SequencerInstruction[T]:
    """Concatenates the given instructions into a single instruction.

    If not all instructions have the same dtype, the result will have the dtype that
    can hold all the values of the instructions.

    Raises:
        ValueError: If there is not at least one instruction provided.
    """

    if len(instructions) == 0:
        raise ValueError("Must provide at least one instruction")
    if not all(
        isinstance(instruction, SequencerInstruction) for instruction in instructions
    ):
        raise TypeError("All instructions must be instances of SequencerInstruction")
    dtype = instructions[0].dtype
    if not all(instruction.dtype == dtype for instruction in instructions):
        result_dtype = np.result_type(
            *[instruction.dtype for instruction in instructions]
        )
        instructions = tuple(
            instruction.as_type(result_dtype) for instruction in instructions
        )
    return _concatenate(*instructions)


def _concatenate[
    T: np.generic
](*instructions: SequencerInstruction[T]) -> SequencerInstruction[T]:
    assert len(instructions) >= 1
    assert all(
        instruction.dtype == instructions[0].dtype for instruction in instructions
    )

    instruction_deque = collections.deque[SequencerInstruction[T]](
        _break_concatenations(instructions)
    )

    useful_instructions: list[SequencerInstruction[T]] = []
    while instruction_deque:
        instruction = instruction_deque.popleft()
        if len(instruction) == 0:
            continue
        if isinstance(instruction, Pattern):
            concatenated_patterns: list[Pattern[T]] = [instruction]
            while instruction_deque and isinstance(instruction_deque[0], Pattern):
                pattern = instruction_deque.popleft()
                assert isinstance(pattern, Pattern)
                concatenated_patterns.append(pattern)
            if len(concatenated_patterns) == 1:
                useful_instructions.append(concatenated_patterns[0])
            else:
                useful_instructions.append(
                    Pattern(
                        numpy.concatenate(
                            [pattern.array for pattern in concatenated_patterns],
                            casting="safe",
                        )
                    )
                )
        else:
            useful_instructions.append(instruction)

    if len(useful_instructions) == 0:
        return empty_like(instructions[0])
    elif len(useful_instructions) == 1:
        return useful_instructions[0]
    else:
        return Concatenated(*useful_instructions)


def _break_concatenations[
    T: np.generic
](instructions: Sequence[SequencerInstruction[T]],) -> list[SequencerInstruction[T]]:
    flat: list[SequencerInstruction[T]] = []
    for instruction in instructions:
        if isinstance(instruction, Concatenated):
            flat.extend(instruction.instructions)
        else:
            flat.append(instruction)
    return flat



from ._html_delegate import HTMLItemDelegate
from ._temporary_widget import temporary_widget
from .block_signals import block_signals
from .qabc import QABC, QABCMeta

__all__ = ["block_signals", "temporary_widget", "QABC", "QABCMeta", "HTMLItemDelegate"]



from typing import Any

import attrs


@attrs.define
class ViewState:
    view_state: Any
    window_geometry: str


@attrs.define
class WorkSpace:
    window_state: str
    window_geometry: str
    view_states: dict[str, ViewState]



from typing import Optional, Any, assert_never

from PySide6.QtCore import QObject, QModelIndex, Qt
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu

from caqtus.types.timelane import AnalogTimeLane, Ramp
from caqtus.types.expression import Expression
from .model import ColoredTimeLaneModel


class AnalogTimeLaneModel(ColoredTimeLaneModel[AnalogTimeLane, None]):
    def __init__(self, name: str, parent: Optional[QObject] = None):
        lane = AnalogTimeLane([Expression("...")])
        super().__init__(name, lane, parent)

    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole
    ):
        if not index.isValid():
            return None
        value = self._lane[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(value, Expression):
                return str(value)
            elif isinstance(value, Ramp):
                return "\u279F"
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.EditRole:
            if isinstance(value, Expression):
                return str(value)
            return None
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter
        else:
            return super().data(index, role)

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        flags = super().flags(index)
        if isinstance(self._lane[index.row()], Ramp):
            flags &= ~Qt.ItemFlag.ItemIsEditable
        return flags

    def setData(
        self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole
    ):
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            start, stop = self._lane.get_bounds(index.row())
            if isinstance(value, str):
                self._lane[start:stop] = Expression(value)
                self.dataChanged.emit(index, index)
                return True
            elif isinstance(value, Ramp):
                self._lane[start:stop] = value
                self.dataChanged.emit(index, index)
                return True
            else:
                raise TypeError(f"Invalid type for value: {type(value)}")
        return False

    def insertRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        return self.insert_value(row, Expression("..."))

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        if not index.isValid():
            return []

        cell_type_menu = QMenu("Cell type")
        value = self._lane[index.row()]
        expr_action = cell_type_menu.addAction("expression")
        if isinstance(value, Expression):
            expr_action.setCheckable(True)
            expr_action.setChecked(True)
        else:
            expr_action.triggered.connect(
                lambda: self.setData(index, "...", Qt.ItemDataRole.EditRole)
            )
        ramp_action = cell_type_menu.addAction("ramp")
        if isinstance(value, Ramp):
            ramp_action.setCheckable(True)
            ramp_action.setChecked(True)
        else:
            ramp_action.triggered.connect(
                lambda: self.setData(index, Ramp(), Qt.ItemDataRole.EditRole)
            )

        return super().get_cell_context_actions(index) + [cell_type_menu]



"""Contains classes for loading data for a shot."""

from ._combinable_importers import CombinableLoader, join
from ._shot_data import ShotImporter, DataImporter
from .load_parameters import LoadShotParameters
from .load_shot_id import LoadShotId
from .load_shot_info import LoadShotTime

__all__ = [
    "CombinableLoader",
    "LoadShotParameters",
    "LoadShotId",
    "LoadShotTime",
    "ShotImporter",
    "join",
    "DataImporter",
]



"""Allows to create user-defined components to use on an experiment."""

from ._experiment import Experiment
from .device_extension import DeviceExtension
from .time_lane_extension import TimeLaneExtension

__all__ = ["Experiment", "DeviceExtension", "TimeLaneExtension"]



#!/usr/bin/python
import re

from PySide6 import QtWidgets, QtCore, QtGui

_NUMB_REGEX = re.compile(r"^((?:\-)*\d+)*([\.,])*(\d+(?:[eE](?:[\-\+])*\d+)*)*")


class _NumberValueMenu(QtWidgets.QMenu):

    mouseMove = QtCore.Signal(object)
    mouseRelease = QtCore.Signal(object)
    stepChange = QtCore.Signal()

    def __init__(self, parent=None):
        super(_NumberValueMenu, self).__init__(parent)
        self.step = 1
        self.steps = []
        self.last_action = None

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    # re-implemented.

    def mousePressEvent(self, event):
        """
        Disabling the mouse press event.
        """
        return

    def mouseReleaseEvent(self, event):
        """
        Additional functionality to emit signal.
        """
        self.mouseRelease.emit(event)
        super(_NumberValueMenu, self).mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):
        """
        Additional functionality to emit step changed signal.
        """
        self.mouseMove.emit(event)
        super(_NumberValueMenu, self).mouseMoveEvent(event)
        action = self.actionAt(event.pos())
        if action:
            if action is not self.last_action:
                self.stepChange.emit()
            self.last_action = action
            self.step = action.step
        elif self.last_action:
            self.setActiveAction(self.last_action)

    def _add_step_action(self, step):
        action = QtGui.QAction(str(step), self)
        action.step = step
        self.addAction(action)

    def set_steps(self, steps):
        self.clear()
        self.steps = steps
        for step in steps:
            self._add_step_action(step)

    def set_data_type(self, data_type):
        if data_type is int:
            new_steps = []
            for step in self.steps:
                if "." not in str(step):
                    new_steps.append(step)
            self.set_steps(new_steps)
        elif data_type is float:
            self.set_steps(self.steps)


class _NumberValueEdit(QtWidgets.QLineEdit):

    value_changed = QtCore.Signal(object)

    def __init__(self, parent=None, data_type=float):
        super(_NumberValueEdit, self).__init__(parent)
        self.setToolTip('"MMB + Drag Left/Right" to change values.')
        self.setText("0")

        self._MMB_STATE = False
        self._previous_x = None
        self._previous_value = None
        self._step = 1
        self._speed = 0.05
        self._data_type = float
        self._min = None
        self._max = None

        self._menu = _NumberValueMenu()
        self._menu.mouseMove.connect(self.mouseMoveEvent)
        self._menu.mouseRelease.connect(self.mouseReleaseEvent)
        self._menu.stepChange.connect(self._reset_previous_x)

        self.editingFinished.connect(self._on_editing_finished)

        self.set_data_type(data_type)

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    # re-implemented

    def mouseMoveEvent(self, event):
        if self._MMB_STATE:
            if self._previous_x is None:
                self._previous_x = event.x()
                self._previous_value = self.get_value()
            else:
                self._step = self._menu.step
                delta = event.x() - self._previous_x
                value = self._previous_value
                value = value + int(delta * self._speed) * self._step
                self.set_value(value)
                self._on_mmb_mouse_move()
        super(_NumberValueEdit, self).mouseMoveEvent(event)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.MouseButton.MiddleButton:
            self._MMB_STATE = True
            self._reset_previous_x()
            self._menu.exec_(QtGui.QCursor.pos())
        super(_NumberValueEdit, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self._menu.close()
        self._MMB_STATE = False
        super(_NumberValueEdit, self).mouseReleaseEvent(event)

    def keyPressEvent(self, event):
        super(_NumberValueEdit, self).keyPressEvent(event)
        if event.key() == QtCore.Qt.Key.Key_Up:
            return
        elif event.key() == QtCore.Qt.Key.Key_Down:
            return

    # private

    def _reset_previous_x(self):
        self._previous_x = None

    def _on_mmb_mouse_move(self):
        self.value_changed.emit(self.get_value())

    def _on_editing_finished(self):
        if self._data_type is float:
            match = _NUMB_REGEX.match(self.text())
            if match:
                val1, point, val2 = match.groups()
                if point:
                    val1 = val1 or "0"
                    val2 = val2 or "0"
                    self.setText(val1 + point + val2)
        self.value_changed.emit(self.get_value())

    def _convert_text(self, text):
        """
        Convert text to int or float.

        Args:
            text (str): input text.

        Returns:
            int or float: converted value.
        """
        match = _NUMB_REGEX.match(text)
        if match:
            val1, _, val2 = match.groups()
            val1 = val1 or "0"
            val2 = val2 or "0"
            value = float(val1 + "." + val2)
        else:
            value = 0.0
        if self._data_type is int:
            value = int(value)
        return value

    # public

    def set_data_type(self, data_type):
        """
        Sets the line edit to either display value in float or int.

        Args:
            data_type(int or float): int or float data type object.
        """
        self._data_type = data_type
        if data_type is int:
            regexp = QtCore.QRegularExpression(r"\d+")
            validator = QtGui.QRegularExpressionValidator(regexp, self)
            steps = [1, 10, 100, 1000]
            self._min = None if self._min is None else int(self._min)
            self._max = None if self._max is None else int(self._max)
        elif data_type is float:
            regexp = QtCore.QRegularExpression(r"\d+[\.,]\d+(?:[eE](?:[\-\+]|)\d+)*")
            validator = QtGui.QRegularExpressionValidator(regexp, self)
            steps = [0.001, 0.01, 0.1, 1]
            self._min = None if self._min is None else float(self._min)
            self._max = None if self._max is None else float(self._max)

        self.setValidator(validator)
        if not self._menu.steps:
            self._menu.set_steps(steps)
        self._menu.set_data_type(data_type)

    def set_steps(self, steps=None):
        """
        Sets the step items in the MMB context menu.

        Args:
            steps (list[int] or list[float]): list of ints or floats.
        """
        step_types = {int: [1, 10, 100, 1000], float: [0.001, 0.01, 0.1, 1]}
        steps = steps or step_types.get(self._data_type)
        self._menu.set_steps(steps)

    def set_min(self, value=None):
        """
        Set the minimum range for the input field.

        Args:
            value (int or float): minimum range value.
        """
        if self._data_type is int:
            self._min = int(value)
        elif self._data_type is float:
            self._min = float(value)
        else:
            self._min = value

    def set_max(self, value=None):
        """
        Set the maximum range for the input field.

        Args:
            value (int or float): maximum range value.
        """
        if self._data_type is int:
            self._max = int(value)
        elif self._data_type is float:
            self._max = float(value)
        else:
            self._max = value

    def get_value(self):
        value = self._convert_text(self.text())
        return value

    def set_value(self, value):
        text = str(value)
        converted = self._convert_text(text)
        current = self.get_value()
        if converted == current:
            return
        point = None
        if isinstance(converted, float):
            point = _NUMB_REGEX.match(str(value)).groups(2)
        if self._min is not None and converted < self._min:
            text = str(self._min)
            if point and point not in text:
                text = str(self._min).replace(".", point)
        if self._max is not None and converted > self._max:
            text = str(self._max)
            if point and point not in text:
                text = text.replace(".", point)
        self.setText(text)


class IntValueEdit(_NumberValueEdit):

    def __init__(self, parent=None):
        super(IntValueEdit, self).__init__(parent, data_type=int)


class FloatValueEdit(_NumberValueEdit):

    def __init__(self, parent=None):
        super(FloatValueEdit, self).__init__(parent, data_type=float)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])

    int_edit = IntValueEdit()
    int_edit.set_steps([1, 10])
    float_edit = FloatValueEdit()

    widget = QtWidgets.QWidget()
    layout = QtWidgets.QVBoxLayout(widget)
    layout.addWidget(int_edit)
    layout.addWidget(float_edit)
    widget.show()

    app.exec_()



#!/usr/bin/python
import re
from collections import OrderedDict

from PySide6 import QtCore, QtWidgets, QtGui

from ..constants import ViewerEnum, ViewerNavEnum


class TabSearchCompleter(QtWidgets.QCompleter):
    """
    QCompleter adapted from:
    https://stackoverflow.com/questions/5129211/qcompleter-custom-completion-rules
    """

    def __init__(self, nodes=None, parent=None):
        super(TabSearchCompleter, self).__init__(nodes, parent)
        self.setCompletionMode(self.PopupCompletion)
        self.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self._local_completion_prefix = ""
        self._using_orig_model = False
        self._source_model = None
        self._filter_model = None

    def splitPath(self, path):
        self._local_completion_prefix = path
        self.updateModel()

        if self._filter_model.rowCount() == 0:
            self._using_orig_model = False
            self._filter_model.setSourceModel(QtCore.QStringListModel([]))
            return []
        return []

    def updateModel(self):
        if not self._using_orig_model:
            self._filter_model.setSourceModel(self._source_model)

        pattern = QtCore.QRegExp(
            self._local_completion_prefix,
            QtCore.Qt.CaseInsensitive,
            QtCore.QRegExp.FixedString,
        )
        self._filter_model.setFilterRegExp(pattern)

    def setModel(self, model):
        self._source_model = model
        self._filter_model = QtCore.QSortFilterProxyModel(self)
        self._filter_model.setSourceModel(self._source_model)
        super(TabSearchCompleter, self).setModel(self._filter_model)
        self._using_orig_model = True


class TabSearchLineEditWidget(QtWidgets.QLineEdit):

    tab_pressed = QtCore.Signal()

    def __init__(self, parent=None):
        super(TabSearchLineEditWidget, self).__init__(parent)
        self.setAttribute(QtCore.Qt.WA_MacShowFocusRect, 0)
        self.setMinimumSize(200, 22)
        # text_color = self.palette().text().color().getRgb()
        text_color = tuple(
            map(lambda i, j: i - j, (255, 255, 255), ViewerEnum.BACKGROUND_COLOR.value)
        )
        selected_color = self.palette().highlight().color().getRgb()
        style_dict = {
            "QLineEdit": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "border": "1px solid rgb({0},{1},{2})".format(*selected_color),
                "border-radius": "3px",
                "padding": "2px 4px",
                "margin": "2px 4px 8px 4px",
                "background": "rgb({0},{1},{2})".format(
                    *ViewerNavEnum.BACKGROUND_COLOR.value
                ),
                "selection-background-color": "rgba({0},{1},{2},200)".format(
                    *selected_color
                ),
            }
        }
        stylesheet = ""
        for css_class, css in style_dict.items():
            style = "{} {{\n".format(css_class)
            for elm_name, elm_val in css.items():
                style += "  {}:{};\n".format(elm_name, elm_val)
            style += "}\n"
            stylesheet += style
        self.setStyleSheet(stylesheet)

    def keyPressEvent(self, event):
        super(TabSearchLineEditWidget, self).keyPressEvent(event)
        if event.key() == QtCore.Qt.Key_Tab:
            self.tab_pressed.emit()


class TabSearchMenuWidget(QtWidgets.QMenu):

    search_submitted = QtCore.Signal(str)

    def __init__(self, node_dict=None):
        super(TabSearchMenuWidget, self).__init__()

        self.line_edit = TabSearchLineEditWidget()
        self.line_edit.tab_pressed.connect(self._close)

        self._node_dict = node_dict or {}
        if self._node_dict:
            self._generate_items_from_node_dict()

        search_widget = QtWidgets.QWidgetAction(self)
        search_widget.setDefaultWidget(self.line_edit)
        self.addAction(search_widget)

        # text_color = self.palette().text().color().getRgb()
        text_color = tuple(
            map(lambda i, j: i - j, (255, 255, 255), ViewerEnum.BACKGROUND_COLOR.value)
        )
        selected_color = self.palette().highlight().color().getRgb()
        style_dict = {
            "QMenu": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "background-color": "rgb({0},{1},{2})".format(
                    *ViewerEnum.BACKGROUND_COLOR.value
                ),
                "border": "1px solid rgba({0},{1},{2},30)".format(*text_color),
                "border-radius": "3px",
            },
            "QMenu::item": {
                "padding": "5px 18px 2px",
                "background-color": "transparent",
            },
            "QMenu::item:selected": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "background-color": "rgba({0},{1},{2},200)".format(*selected_color),
            },
            "QMenu::separator": {
                "height": "1px",
                "background": "rgba({0},{1},{2}, 50)".format(*text_color),
                "margin": "4px 8px",
            },
        }
        self._menu_stylesheet = ""
        for css_class, css in style_dict.items():
            style = "{} {{\n".format(css_class)
            for elm_name, elm_val in css.items():
                style += "  {}:{};\n".format(elm_name, elm_val)
            style += "}\n"
            self._menu_stylesheet += style
        self.setStyleSheet(self._menu_stylesheet)

        self._actions = {}
        self._menus = {}
        self._searched_actions = []

        self._block_submit = False

        self.rebuild = False

        self._wire_signals()

    def __repr__(self):
        return "<{} at {}>".format(self.__class__.__name__, hex(id(self)))

    def keyPressEvent(self, event):
        super(TabSearchMenuWidget, self).keyPressEvent(event)
        self.line_edit.keyPressEvent(event)

    @staticmethod
    def _fuzzy_finder(key, collection):
        suggestions = []
        pattern = ".*?".join(key.lower())
        regex = re.compile(pattern)
        for item in collection:
            match = regex.search(item.lower())
            if match:
                suggestions.append((len(match.group()), match.start(), item))

        return [x for _, _, x in sorted(suggestions)]

    def _wire_signals(self):
        self.line_edit.returnPressed.connect(self._on_search_submitted)
        self.line_edit.textChanged.connect(self._on_text_changed)

    def _on_text_changed(self, text):
        self._clear_actions()

        if not text:
            self._set_menu_visible(True)
            return

        self._set_menu_visible(False)

        action_names = self._fuzzy_finder(text, self._actions.keys())

        self._searched_actions = [self._actions[name] for name in action_names]
        self.addActions(self._searched_actions)

        if self._searched_actions:
            self.setActiveAction(self._searched_actions[0])

    def _clear_actions(self):
        for action in self._searched_actions:
            self.removeAction(action)
            action.triggered.connect(self._on_search_submitted)
        del self._searched_actions[:]

    def _set_menu_visible(self, visible):
        for menu in self._menus.values():
            menu.menuAction().setVisible(visible)

    def _close(self):
        self._set_menu_visible(False)
        self.setVisible(False)
        self.menuAction().setVisible(False)
        self._block_submit = True

    def _show(self):
        self.line_edit.setText("")
        self.line_edit.setFocus()
        self._set_menu_visible(True)
        self._block_submit = False
        self.exec_(QtGui.QCursor.pos())

    def _on_search_submitted(self):
        if not self._block_submit:
            action = self.sender()
            if type(action) is not QtGui.QAction:
                if len(self._searched_actions) > 0:
                    action = self._searched_actions[0]
                else:
                    self._close()
                    return

            text = action.text()
            node_type = self._node_dict.get(text)
            if node_type:
                self.search_submitted.emit(node_type)

        self._close()

    def build_menu_tree(self):
        node_types = sorted(self._node_dict.values())
        node_names = sorted(self._node_dict.keys())
        menu_tree = OrderedDict()

        max_depth = 0
        for node_type in node_types:
            trees = ".".join(node_type.split(".")[:-1]).split("::")
            for depth, menu_name in enumerate(trees):
                new_menu = None
                menu_path = "::".join(trees[: depth + 1])
                if depth in menu_tree.keys():
                    if menu_name not in menu_tree[depth].keys():
                        new_menu = QtWidgets.QMenu(menu_name)
                        new_menu.keyPressEvent = self.keyPressEvent
                        new_menu.setStyleSheet(self._menu_stylesheet)
                        menu_tree[depth][menu_path] = new_menu
                else:
                    new_menu = QtWidgets.QMenu(menu_name)
                    new_menu.setStyleSheet(self._menu_stylesheet)
                    menu_tree[depth] = {menu_path: new_menu}
                if depth > 0 and new_menu:
                    new_menu.parentPath = "::".join(trees[:depth])

                max_depth = max(max_depth, depth)

        for i in range(max_depth + 1):
            menus = menu_tree[i]
            for menu_path, menu in menus.items():
                self._menus[menu_path] = menu
                if i == 0:
                    self.addMenu(menu)
                else:
                    parent_menu = self._menus[menu.parentPath]
                    parent_menu.addMenu(menu)

        for name in node_names:
            action = QtGui.QAction(name, self)
            action.setText(name)
            action.triggered.connect(self._on_search_submitted)
            self._actions[name] = action

            menu_name = self._node_dict[name]
            menu_path = ".".join(menu_name.split(".")[:-1])

            if menu_path in self._menus.keys():
                self._menus[menu_path].addAction(action)
            else:
                self.addAction(action)

    def set_nodes(self, node_dict=None):
        if not self._node_dict or self.rebuild:
            self._node_dict.clear()
            self._clear_actions()
            self._set_menu_visible(False)
            for menu in self._menus.values():
                self.removeAction(menu.menuAction())
            self._actions.clear()
            self._menus.clear()
            for name, node_types in node_dict.items():
                if len(node_types) == 1:
                    self._node_dict[name] = node_types[0]
                    continue
                for node_id in node_types:
                    self._node_dict["{} ({})".format(name, node_id)] = node_id
            self.build_menu_tree()
            self.rebuild = False

        self._show()



from __future__ import annotations

import datetime
import re
from typing import Self, Optional, TYPE_CHECKING, Iterable

from ._return_or_raise import unwrap

if TYPE_CHECKING:
    from ._experiment_session import ExperimentSession

_PATH_SEPARATOR = "\\"
_CHARACTER_SET = (
    "["
    "\\w\\d"
    "\\_\\.\\,\\;\\*\\'\\\"\\`"
    "\\(\\)\\[\\]\\{\\}"
    "\\+\\-\\*\\/\\="
    "\\!\\@\\#\\$\\%\\^\\&\\~\\<\\>\\?\\|"
    "]"
)
_PATH_NAME = f"{_CHARACTER_SET}+(?:{_CHARACTER_SET}| )*"
_PATH_NAME_REGEX = re.compile(f"^{_PATH_NAME}$")
_PATH_REGEX = re.compile(f"^\\{_PATH_SEPARATOR}|(\\{_PATH_SEPARATOR}{_PATH_NAME})+$")


class PureSequencePath:
    r"""Represent a path in the sequence hierarchy.

    A path is a string of names separated by a single backslash.
    For example, "\\foo\\bar" is a path with two parts, "foo" and "bar".
    The root path is the single backslash "\\".

    All methods of this class are *pure* in the sense that they do not interact with
    the storage system and only manipulate the path string.
    """

    __slots__ = ("_parts", "_str")

    def __init__(self, path: PureSequencePath | str):
        """

        Raises:
            InvalidPathFormatError: If the path has an invalid format.
        """
        if isinstance(path, str):
            self._parts = self._convert_to_parts(path)
            self._str = path
        elif isinstance(path, PureSequencePath):
            self._parts = path.parts
            self._str = path._str
        else:
            raise TypeError(f"Invalid type for path: {type(path)}")

    @property
    def parts(self) -> tuple[str, ...]:
        r"""Return the parts of the path.

        The parts are the names that make up the path.

        The root path has no parts and this attribute is an empty tuple for it.

        Example:
            >>> path = PureSequencePath(r"\foo\bar")
            >>> path.parts
            ('foo', 'bar')
        """

        return self._parts

    @property
    def parent(self) -> Optional[PureSequencePath]:
        r"""Return the parent of this path.

        Returns:
            The parent of this path, or :data:`None` if this path is the root path.

        Example:
            >>> path = PureSequencePath(r"\foo\bar")
            >>> path.parent
            PureSequencePath("\\foo")
        """

        if self.is_root():
            return None
        else:
            return PureSequencePath.from_parts(self._parts[:-1])

    def get_ancestors(self) -> Iterable[Self]:
        r"""Return the ancestors of this path.

        Returns:
            All the paths that are above this path in the hierarchy, ordered from the
            current path to the root, both included.

            For the root path, the result will only contain the root path.

        Example:
            >>> path = PureSequencePath(r"\foo\bar\baz")
            >>> list(path.get_ancestors())
            [
                PureSequencePath("\\foo\\bar\\baz"),
                PureSequencePath("\\foo\\bar"),
                PureSequencePath("\\foo"),
                PureSequencePath("\\"),
            ]
        """

        current = self
        yield current
        while parent := current.parent:
            current = parent
            yield current

    @property
    def name(self) -> Optional[str]:
        r"""Return the last part of the path.

        The root path has no name and this attribute is None for it.

        Example:
            >>> path = PureSequencePath(r"\foo\bar")
            >>> path.name
            'bar'
        """

        if self.is_root():
            return None
        else:
            return self._parts[-1]

    def __str__(self) -> str:
        return self._str

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._str!r})"

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._str == other._str
        else:
            return NotImplemented

    def is_root(self) -> bool:
        """Check if the path is the root path."""

        return len(self._parts) == 0

    def __truediv__(self, other) -> Self:
        """Add a name to the path.

        Example:
            >>> path = PureSequencePath.root()
            >>> path / "foo"
            PureSequencePath("\\foo")

        Raises:
            InvalidPathFormatError: If the name is not valid.
        """

        if isinstance(other, str):
            if not re.match(_PATH_NAME, other):
                raise InvalidPathFormatError("Invalid name format")
            if self.is_root():
                return type(self)(f"{self._str}{other}")
            else:
                return type(self)(f"{self._str}{_PATH_SEPARATOR}{other}")
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._str)

    @classmethod
    def root(cls) -> PureSequencePath:
        r"""Returns the root path.

        The root path is represented by the single backslash character "\\".
        """

        return cls(cls._separator())

    @classmethod
    def _separator(cls) -> str:
        """Returns the character used to separate path names."""

        return _PATH_SEPARATOR

    @classmethod
    def is_valid_name(cls, name: str) -> bool:
        """Check if a string is a valid name."""

        return bool(_PATH_NAME_REGEX.match(name))

    @classmethod
    def _convert_to_parts(cls, path: str) -> tuple[str, ...]:
        if cls.is_valid_path(path):
            if path == _PATH_SEPARATOR:
                return tuple()
            else:
                return tuple(path.split(_PATH_SEPARATOR)[1:])
        else:
            raise InvalidPathFormatError(f"Invalid path: {path}")

    @classmethod
    def from_parts(cls, parts: Iterable[str]) -> PureSequencePath:
        """Create a path from its parts.

        Raises:
            InvalidPathFormatError: If one of the parts is not a valid name.
        """

        return PureSequencePath(_PATH_SEPARATOR + _PATH_SEPARATOR.join(parts))

    @classmethod
    def is_valid_path(cls, path: str) -> bool:
        """Check if a string is a valid path."""

        return bool(_PATH_REGEX.match(path))


class BoundSequencePath(PureSequencePath):
    __slots__ = ("_session",)

    def __init__(self, path: PureSequencePath | str, session: "ExperimentSession"):
        super().__init__(path)
        self._session = session

    @property
    def parent(self) -> Optional[Self]:
        p = super().parent
        if p is None:
            return None
        else:
            return BoundSequencePath(p, self._session)

    def get_ancestors(self) -> Iterable[Self]:
        ancestors = super().get_ancestors()
        return (BoundSequencePath(a, self._session) for a in ancestors)

    def rebind(self, new_session: "ExperimentSession"):
        self._session = new_session

    def exists(self) -> bool:
        return self._session.paths.does_path_exists(self)

    def create(self) -> list[BoundSequencePath]:
        """Create the path and all its ancestors if they don't exist.

        Return:
            A list of paths that were created if they didn't exist.

        Raises:
            PathIsSequenceError: If an ancestor exists and is a sequence.
        """

        result = self._session.paths.create_path(self)
        return [BoundSequencePath(path, self._session) for path in unwrap(result)]

    def get_children(self) -> set[BoundSequencePath]:
        """Return the direct descendants of this path.

        Returns:
            A set of the direct descendants of this path.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
            PathIsSequenceError: If the path is a sequence.
        """

        result = self._session.paths.get_children(self)
        return {BoundSequencePath(path, self._session) for path in unwrap(result)}

    def get_descendants(self) -> set[BoundSequencePath]:
        """Return the descendants of this path.

        Returns:
            A set of the descendants of this path.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
            PathIsSequenceError: If the path is a sequence.
        """

        from ._sequence_collection import PathIsSequenceError

        descendants = set()
        for child in self.get_children():
            descendants.add(child)
            try:
                descendants.update(child.get_descendants())
            except PathIsSequenceError:
                pass
        return descendants

    def delete(self, delete_sequences: bool = False):
        """Delete the path and all its children if they exist.

        Warnings:
            If delete_sequences is True, all sequences in the path will be deleted.

        Raises:
            PathIsSequenceError: If the path or one of its children is a sequence and
            delete_sequence is False
        """

        self._session.paths.delete_path(self, delete_sequences)

    def get_creation_date(
        self, experiment_session: "ExperimentSession"
    ) -> datetime.datetime:
        """Get the creation date of the path.

        Returns:
            The date at which the path was created.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
        """

        result = experiment_session.paths.get_path_creation_date(self)
        return unwrap(result)

    @property
    def session(self) -> "ExperimentSession":
        return self._session


class InvalidPathFormatError(ValueError):
    """Raised when a path has an invalid format."""

    pass



import contextlib
from typing import Protocol

from caqtus.types.parameter import ParameterNamespace
from .path_hierarchy import AsyncPathHierarchy
from .sequence_collection import AsyncSequenceCollection


class AsyncExperimentSession(
    contextlib.AbstractAsyncContextManager["AsyncExperimentSession"], Protocol
):
    """Asynchronous version of ExperimentSession.

    For a detailed description of the methods, see ExperimentSession.
    This session cannot be used in several tasks concurrently, it must be used
    sequentially.
    """

    paths: AsyncPathHierarchy
    sequences: AsyncSequenceCollection

    async def get_global_parameters(self) -> ParameterNamespace: ...

    async def set_global_parameters(self, parameters: ParameterNamespace) -> None: ...



from collections.abc import Iterable
from itertools import tee
from typing import TypeVar

T = TypeVar("T")


def pairwise(iterable: Iterable[T]) -> Iterable[tuple[T, T]]:
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)



from typing import Optional

from PySide6.QtWidgets import QDialog, QWidget

from analyza.loading.importers import ImageLoader
from device.name import DeviceName
from image_types import ImageLabel
from viewer.single_shot_viewers import ImageViewer
from .add_image_dialog_ui import Ui_ImageDialog


def create_image_viewer(parent: Optional[QWidget]) -> Optional[tuple[str, ImageViewer]]:
    dialog = ViewerDialog(parent)

    if dialog.exec() == QDialog.DialogCode.Accepted:
        return (
            dialog.get_title(),
            ImageViewer(
                importer=ImageLoader(
                    dialog.get_camera_name(),
                    dialog.get_picture_name(),
                ),
            ),
        )
    return None


class ViewerDialog(QDialog, Ui_ImageDialog):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setupUi(self)

    def get_title(self) -> str:
        return self._window_title.text()

    def get_camera_name(self) -> DeviceName:
        return DeviceName(self._camera_line_edit.text())

    def get_picture_name(self) -> ImageLabel:
        return ImageLabel(self._picture_line_edit.text())



"""This module defines the configuration used to compute the output of a sequencer
channel.

A channel can typically output a constant value, the values of a lane, a trigger for
another device, or a functional combination of these.

The union type `ChannelOutput` is used to represent the different possible outputs of a
channel.
Each possible type of output is represented by a different class.
An output class is a high-level description of what should be outputted by a channel.
The classes defined are only declarative and do not contain any logic to compute the
output.
For more information on how the output is evaluated, see
:mod:`core.compilation.sequencer_parameter_compiler`.
"""

from __future__ import annotations

import abc
from collections.abc import Mapping
from typing import Optional, Any

import attrs
import numpy as np

from caqtus.device.sequencer.instructions import SequencerInstruction
from caqtus.shot_compilation import ShotContext
from caqtus.types.units import Unit, dimensionless
from caqtus.types.units.base import is_in_base_units
from caqtus.types.variable_name import DottedVariableName
from .._time_step import TimeStep


@attrs.frozen
class DimensionedSeries[T: (np.number, np.bool_)]:
    """Represents a series of value to output on a channel with their units.

    Parameters:
        values: The sequence of values to output.
        units: The units in which the values are expressed.
            The units must be expressed in the base units of the registry.
            If the values are dimensionless, the units must be `None`.
    """

    values: SequencerInstruction[T]
    units: Optional[Unit] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(Unit))
    )

    @units.validator  # type: ignore
    def _validate_units(self, _, units: Optional[Unit]):
        if units is not None:
            if not is_in_base_units(units):
                raise ValueError(
                    f"Unit {units} is not expressed in the base units of the registry."
                )
            if units.is_compatible_with(dimensionless):
                raise ValueError(f"Unit {units} is dimensionless and must be None.")


@attrs.define
class ChannelOutput(abc.ABC):
    """Defines what should be outputted by a channel.

    Subclasses of this class should have specific attributes that define how to
    evaluate the series of values to output on the channel.

    These attributes can contain other :class:`ChannelOutput` instances, which allows
    to recursively combine transformations.
    """

    @abc.abstractmethod
    def evaluate[
        T: (np.number, np.bool_)
    ](
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ) -> DimensionedSeries[T]:
        """Evaluate the output of a channel with the required parameters.

        Args:
            required_time_step: The time step in which to evaluate the output, in ns.
            prepend: The number of time steps to add at the beginning of the output.
            append: The number of time steps to add at the end of the output.
            shot_context: Contains information about the current to evaluate the output.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        raise NotImplementedError



from datetime import datetime
from typing import Any, Callable, TypeVar

from cattrs.converters import Converter
from cattrs.preconf.json import make_converter as make_json_converter, JsonConverter
from cattrs.preconf.pyyaml import make_converter as make_yaml_converter

unstruct_collection_overrides = {tuple: tuple}

T = TypeVar("T")

converters = {
    "json": make_json_converter(
        unstruct_collection_overrides=unstruct_collection_overrides
    ),
    "yaml": make_yaml_converter(
        unstruct_collection_overrides=unstruct_collection_overrides
    ),
    "unconfigured": Converter(
        unstruct_collection_overrides=unstruct_collection_overrides
    ),
}


def unstructure(obj: Any, unstructure_as: Any = None):
    return converters["unconfigured"].unstructure(obj, unstructure_as=unstructure_as)


def structure(obj: Any, cls: Any = None):
    return converters["unconfigured"].structure(obj, cls)


def register_unstructure_hook(cls: Any, hook: Callable[[Any], Any]) -> None:
    """Register a class-to-primitive converter function for a class.

    The converter function should take an instance of the class and return
    its Python equivalent.
    """

    for converter in converters.values():
        converter.register_unstructure_hook(cls, hook)


def register_structure_hook(cls: Any, func: Callable[[Any, type[T]], T]) -> None:
    """Register a primitive-to-class converter function for a type.

    The converter function should take two arguments:
      * a Python object to be converted,
      * the type to convert to

    and return the instance of the class. The type may seem redundant, but
    is sometimes needed (for example, when dealing with generic classes).
    """

    for converter in converters.values():
        converter.register_structure_hook(cls, func)


# datetime serialization


def unstructure_datetime(obj: datetime) -> str:
    return obj.isoformat()


register_unstructure_hook(datetime, unstructure_datetime)


def structure_datetime(serialized: str, _) -> datetime:
    return datetime.fromisoformat(serialized)


register_structure_hook(datetime, structure_datetime)


def to_json(obj: Any, unstructure_as: Any = None) -> str:
    converter: JsonConverter = converters["json"]  # type: ignore
    return converter.dumps(obj, unstructure_as=unstructure_as, indent=2)


def from_json(serialized: str, cls: type[T]) -> T:
    converter: JsonConverter = converters["json"]  # type: ignore
    return converter.loads(serialized, cls)


def copy_converter() -> Converter:
    """Return a copy of a serialization converter with common hooks registered."""

    return converters["json"].copy()



from collections.abc import Callable
from typing import TypeVar, Optional, NewType

from caqtus.utils import serialization
from caqtus.utils.serialization import JSON
from ._protocol import TimeLaneSerializerProtocol
from ..timelane import TimeLane, TimeLanes
from ...expression import Expression

L = TypeVar("L", bound=TimeLane)

Tag = NewType("Tag", str)
Dumper = Callable[[L], JSON]
Loader = Callable[[JSON], L]


class TimeLaneSerializer(TimeLaneSerializerProtocol):
    def __init__(self):
        self.dumpers: dict[type, tuple[Dumper, Tag]] = {}
        self.loaders: dict[Tag, Loader] = {}

    def register_time_lane(
        self,
        lane_type: type[L],
        dumper: Dumper[L],
        loader: Loader[L],
        type_tag: Optional[str] = None,
    ) -> None:
        if type_tag is None:
            tag = Tag(lane_type.__qualname__)
        else:
            tag = Tag(type_tag)
        self.dumpers[lane_type] = (dumper, tag)
        self.loaders[tag] = loader

    def dump(self, lane: TimeLane) -> JSON:
        dumper, tag = self.dumpers[type(lane)]
        content = dumper(lane)
        if "type" in content:
            raise ValueError("The content already has a type tag.")
        content["type"] = tag
        return content

    def load(self, data: JSON) -> TimeLane:
        tag = data["type"]
        loader = self.loaders[tag]
        return loader(data)

    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> serialization.JSON:
        return dict(
            step_names=serialization.converters["json"].unstructure(
                time_lanes.step_names, list[str]
            ),
            step_durations=serialization.converters["json"].unstructure(
                time_lanes.step_durations, list[Expression]
            ),
            lanes={
                lane: self.dump(time_lane)
                for lane, time_lane in time_lanes.lanes.items()
            },
        )

    def structure_time_lanes(self, content: JSON) -> TimeLanes:
        return TimeLanes(
            step_names=serialization.converters["json"].structure(
                content["step_names"], list[str]
            ),
            step_durations=serialization.converters["json"].structure(
                content["step_durations"], list[Expression]
            ),
            lanes={
                lane: self.load(time_lane_content)
                for lane, time_lane_content in content["lanes"].items()
            },
        )


def default_dumper(lane) -> JSON:
    raise NotImplementedError(f"Unsupported type {type(lane)}")



from typing import assert_never

import attrs
from caqtus.types.expression import Expression
from caqtus.utils import serialization

from .timelane import TimeLane


@attrs.define(init=False, eq=False, repr=False)
class DigitalTimeLane(TimeLane[bool | Expression]):
    """A time lane that represents digital values over time.

    The values of a digital time lane can be a boolean: True or False, or a
    placeholder expression.
    The placeholder expression is used to represent a value that is not known
    at the time of creation of the time lane and will be evaluated at a later
    time.

    Examples:

    .. code-block:: python

       from caqtus.session.shot import DigitalTimeLane
       from caqtus.types.expression import Expression

       # Creates a digital time lane with known values
       lane = DigitalTimeLane([True] * 3 + [False, True])

       # Creates a digital time lane with a placeholder expression
       lane = DigitalTimeLane([Expression("x"), False, True])
    """

    pass


def union_unstructure(union: bool | Expression):
    if isinstance(union, bool):
        return union
    elif isinstance(union, Expression):
        return union.body
    else:
        assert_never(union)


def union_structure(data, _) -> bool | Expression:
    if isinstance(data, bool):
        return data
    elif isinstance(data, str):
        return Expression(data)
    else:
        raise ValueError(f"Invalid union value: {data}")


serialization.register_structure_hook(bool | Expression, union_structure)
serialization.register_unstructure_hook(bool | Expression, union_unstructure)


def unstructure_hook(lane: DigitalTimeLane):
    return {
        "spanned_values": serialization.unstructure(
            lane._spanned_values, list[tuple[bool | Expression, int]]
        )
    }


def structure_hook(data, _) -> DigitalTimeLane:
    structured = serialization.structure(
        data["spanned_values"], list[tuple[bool | Expression, int]]
    )
    return DigitalTimeLane.from_spanned_values(structured)


serialization.register_structure_hook(DigitalTimeLane, structure_hook)
serialization.register_unstructure_hook(DigitalTimeLane, unstructure_hook)



import abc
import datetime
from collections.abc import Mapping
from typing import Protocol

from returns.result import Result

from caqtus.types.data import DataLabel, Data
from caqtus.types.iteration import IterationConfiguration
from caqtus.types.parameter import Parameter, ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from .._path import PureSequencePath
from .._path_hierarchy import PathNotFoundError
from .._sequence_collection import PathIsNotSequenceError, SequenceStats, PureShot
from .._state import State


class AsyncSequenceCollection(Protocol):
    @abc.abstractmethod
    async def is_sequence(
        self, path: PureSequencePath
    ) -> Result[bool, PathNotFoundError]:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_stats(
        self, path: PureSequencePath
    ) -> Result[SequenceStats, PathNotFoundError | PathIsNotSequenceError]:
        raise NotImplementedError

    async def get_state(
        self, path: PureSequencePath
    ) -> Result[State, PathNotFoundError | PathIsNotSequenceError]:

        return (await self.get_stats(path)).map(lambda stats: stats.state)

    @abc.abstractmethod
    async def get_iteration_configuration(
        self, path: PureSequencePath
    ) -> IterationConfiguration:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_time_lanes(self, path: PureSequencePath) -> TimeLanes:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_global_parameters(self, path: PureSequencePath) -> ParameterNamespace:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_shots(
        self, path: PureSequencePath
    ) -> Result[list[PureShot], PathNotFoundError | PathIsNotSequenceError]:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_shot_parameters(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DottedVariableName, Parameter]:

        raise NotImplementedError

    @abc.abstractmethod
    async def get_all_shot_data(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DataLabel, Data]:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_shot_data_by_label(
        self, path: PureSequencePath, shot_index: int, data_label: DataLabel
    ) -> Data:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_shot_start_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_shot_end_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        raise NotImplementedError



#!/usr/bin/python
import json
from collections import defaultdict

from ..constants import LayoutDirectionEnum, NodePropWidgetEnum, PipeLayoutEnum
from ..errors import NodePropertyError


class PortModel(object):
    """
    Data dump for a port object.
    """

    def __init__(self, node):
        self.node = node
        self.type_ = ""
        self.name = "port"
        self.display_name = True
        self.multi_connection = False
        self.visible = True
        self.locked = False
        self.connected_ports = defaultdict(list)

    def __repr__(self):
        return "<{}('{}') object at {}>".format(
            self.__class__.__name__, self.name, hex(id(self))
        )

    @property
    def to_dict(self):
        """
        serialize model information to a dictionary.

        Returns:
            dict: node port dictionary eg.
                {
                    'type': 'in',
                    'name': 'port',
                    'display_name': True,
                    'multi_connection': False,
                    'visible': True,
                    'locked': False,
                    'connected_ports': {<node_id>: [<port_name>, <port_name>]}
                }
        """
        props = self.__dict__.copy()
        props.pop("node")
        props["connected_ports"] = dict(props.pop("connected_ports"))
        return props


class NodeModel(object):
    """
    Data dump for a node object.
    """

    def __init__(self):
        self.type_ = None
        self.id = hex(id(self))
        self.icon = None
        self.name = "node"
        self.color = (13, 18, 23, 255)
        self.border_color = (74, 84, 85, 255)
        self.text_color = (255, 255, 255, 180)
        self.disabled = False
        self.selected = False
        self.visible = True
        self.width = 100.0
        self.height = 80.0
        self.pos = [0.0, 0.0]
        self.layout_direction = LayoutDirectionEnum.HORIZONTAL.value

        # BaseNode attrs.
        self.inputs = {}
        self.outputs = {}
        self.port_deletion_allowed = False

        # GroupNode attrs.
        self.subgraph_session = {}

        # Custom
        self._custom_prop = {}

        # node graph model set at node added time.
        self._graph_model = None

        # store the property attributes.
        # (deleted when node is added to the graph)
        self._TEMP_property_attrs = {}

        # temp store the property widget types.
        # (deleted when node is added to the graph)
        self._TEMP_property_widget_types = {
            "type_": NodePropWidgetEnum.QLABEL.value,
            "id": NodePropWidgetEnum.QLABEL.value,
            "icon": NodePropWidgetEnum.HIDDEN.value,
            "name": NodePropWidgetEnum.QLINE_EDIT.value,
            "color": NodePropWidgetEnum.COLOR_PICKER.value,
            "border_color": NodePropWidgetEnum.COLOR_PICKER.value,
            "text_color": NodePropWidgetEnum.COLOR_PICKER.value,
            "disabled": NodePropWidgetEnum.QCHECK_BOX.value,
            "selected": NodePropWidgetEnum.HIDDEN.value,
            "width": NodePropWidgetEnum.HIDDEN.value,
            "height": NodePropWidgetEnum.HIDDEN.value,
            "pos": NodePropWidgetEnum.HIDDEN.value,
            "layout_direction": NodePropWidgetEnum.HIDDEN.value,
            "inputs": NodePropWidgetEnum.HIDDEN.value,
            "outputs": NodePropWidgetEnum.HIDDEN.value,
        }

        # temp store connection constrains.
        # (deleted when node is added to the graph)
        self._TEMP_accept_connection_types = {}
        self._TEMP_reject_connection_types = {}

    def __repr__(self):
        return "<{}('{}') object at {}>".format(
            self.__class__.__name__, self.name, self.id
        )

    def add_property(
        self,
        name,
        value,
        items=None,
        range=None,
        widget_type=None,
        widget_tooltip=None,
        tab=None,
    ):
        """
        add custom property or raises an error if the property name is already
        taken.

        Args:
            name (str): name of the property.
            value (object): data.
            items (list[str]): items used by widget type NODE_PROP_QCOMBO.
            range (tuple): min, max values used by NODE_PROP_SLIDER.
            widget_type (int): widget type flag.
            widget_tooltip (str): custom tooltip for the property widget.
            tab (str): widget tab name.
        """
        widget_type = widget_type or NodePropWidgetEnum.HIDDEN.value
        tab = tab or "Properties"

        if name in self.properties.keys():
            raise NodePropertyError('"{}" reserved for default property.'.format(name))
        if name in self._custom_prop.keys():
            raise NodePropertyError('"{}" property already exists.'.format(name))

        self._custom_prop[name] = value

        if self._graph_model is None:
            self._TEMP_property_widget_types[name] = widget_type
            self._TEMP_property_attrs[name] = {"tab": tab}
            if items:
                self._TEMP_property_attrs[name]["items"] = items
            if range:
                self._TEMP_property_attrs[name]["range"] = range
            if widget_tooltip:
                self._TEMP_property_attrs[name]["tooltip"] = widget_tooltip

        else:
            attrs = {self.type_: {name: {"widget_type": widget_type, "tab": tab}}}
            if items:
                attrs[self.type_][name]["items"] = items
            if range:
                attrs[self.type_][name]["range"] = range
            if widget_tooltip:
                attrs[self.type_][name]["tooltip"] = widget_tooltip
            self._graph_model.set_node_common_properties(attrs)

    def set_property(self, name, value):
        """
        Args:
            name (str): property name.
            value (object): property value.
        """
        if name in self.properties.keys():
            setattr(self, name, value)
        elif name in self._custom_prop.keys():
            self._custom_prop[name] = value
        else:
            raise NodePropertyError('No property "{}"'.format(name))

    def get_property(self, name):
        """
        Args:
            name (str): property name.

        Returns:
            object: property value.
        """
        if name in self.properties.keys():
            return self.properties[name]
        return self._custom_prop.get(name)

    def is_custom_property(self, name):
        """
        Args:
            name (str): property name.

        Returns:
            bool: true if custom property.
        """
        return name in self._custom_prop

    def get_widget_type(self, name):
        """
        Args:
            name (str): property name.

        Returns:
            int: node property widget type.
        """
        model = self._graph_model
        if model is None:
            return self._TEMP_property_widget_types.get(name)
        return model.get_node_common_properties(self.type_)[name]["widget_type"]

    def get_tab_name(self, name):
        """
        Args:
            name (str): property name.

        Returns:
            str: name of the tab for the properties bin.
        """
        model = self._graph_model
        if model is None:
            attrs = self._TEMP_property_attrs.get(name)
            if attrs:
                return attrs[name].get("tab")
            return
        return model.get_node_common_properties(self.type_)[name]["tab"]

    def add_port_accept_connection_type(
        self, port_name, port_type, node_type, accept_pname, accept_ptype, accept_ntype
    ):
        """
        Convenience function for adding to the "accept_connection_types" dict.
        If the node graph model is unavailable yet then we store it to a
        temp var that gets deleted.

        Args:
            port_name (str): current port name.
            port_type (str): current port type.
            node_type (str): current port node type.
            accept_pname (str):port name to accept.
            accept_ptype (str): port type accept.
            accept_ntype (str):port node type to accept.
        """
        model = self._graph_model
        if model:
            model.add_port_accept_connection_type(
                port_name,
                port_type,
                node_type,
                accept_pname,
                accept_ptype,
                accept_ntype,
            )
            return

        connection_data = self._TEMP_accept_connection_types
        keys = [node_type, port_type, port_name, accept_ntype]
        for key in keys:
            if key not in connection_data.keys():
                connection_data[key] = {}
            connection_data = connection_data[key]

        if accept_ptype not in connection_data:
            connection_data[accept_ptype] = set([accept_pname])
        else:
            connection_data[accept_ptype].add(accept_pname)

    def add_port_reject_connection_type(
        self, port_name, port_type, node_type, reject_pname, reject_ptype, reject_ntype
    ):
        """
        Convenience function for adding to the "reject_connection_types" dict.
        If the node graph model is unavailable yet then we store it to a
        temp var that gets deleted.

        Args:
            port_name (str): current port name.
            port_type (str): current port type.
            node_type (str): current port node type.
            reject_pname:
            reject_ptype:
            reject_ntype:

        Returns:

        """
        model = self._graph_model
        if model:
            model.add_port_reject_connection_type(
                port_name,
                port_type,
                node_type,
                reject_pname,
                reject_ptype,
                reject_ntype,
            )
            return

        connection_data = self._TEMP_reject_connection_types
        keys = [node_type, port_type, port_name, reject_ntype]
        for key in keys:
            if key not in connection_data.keys():
                connection_data[key] = {}
            connection_data = connection_data[key]

        if reject_ptype not in connection_data:
            connection_data[reject_ptype] = set([reject_pname])
        else:
            connection_data[reject_ptype].add(reject_pname)

    @property
    def properties(self):
        """
        return all default node properties.

        Returns:
            dict: default node properties.
        """
        props = self.__dict__.copy()
        exclude = [
            "_custom_prop",
            "_graph_model",
            "_TEMP_property_attrs",
            "_TEMP_property_widget_types",
        ]
        [props.pop(i) for i in exclude if i in props.keys()]
        return props

    @property
    def custom_properties(self):
        """
        return all custom properties specified by the user.

        Returns:
            dict: user defined properties.
        """
        return self._custom_prop

    @property
    def to_dict(self):
        """
        serialize model information to a dictionary.

        Returns:
            dict: node id as the key and properties as the values eg.
                {'0x106cf75a8': {
                    'name': 'foo node',
                    'color': (48, 58, 69, 255),
                    'border_color': (85, 100, 100, 255),
                    'text_color': (255, 255, 255, 180),
                    'type_': 'io.github.jchanvfx.FooNode',
                    'selected': False,
                    'disabled': False,
                    'visible': True,
                    'width': 0.0,
                    'height: 0.0,
                    'pos': (0.0, 0.0),
                    'layout_direction': 0,
                    'custom': {},
                    'inputs': {
                        <port_name>: {<node_id>: [<port_name>, <port_name>]}
                    },
                    'outputs': {
                        <port_name>: {<node_id>: [<port_name>, <port_name>]}
                    },
                    'input_ports': [<port_name>, <port_name>],
                    'output_ports': [<port_name>, <port_name>],
                    },
                    subgraph_session: <sub graph session data>
                }
        """
        node_dict = self.__dict__.copy()
        node_id = node_dict.pop("id")

        inputs = {}
        outputs = {}
        input_ports = []
        output_ports = []
        for name, model in node_dict.pop("inputs").items():
            if self.port_deletion_allowed:
                input_ports.append(
                    {
                        "name": name,
                        "multi_connection": model.multi_connection,
                        "display_name": model.display_name,
                    }
                )
            connected_ports = model.to_dict["connected_ports"]
            if connected_ports:
                inputs[name] = connected_ports
        for name, model in node_dict.pop("outputs").items():
            if self.port_deletion_allowed:
                output_ports.append(
                    {
                        "name": name,
                        "multi_connection": model.multi_connection,
                        "display_name": model.display_name,
                    }
                )
            connected_ports = model.to_dict["connected_ports"]
            if connected_ports:
                outputs[name] = connected_ports
        if inputs:
            node_dict["inputs"] = inputs
        if outputs:
            node_dict["outputs"] = outputs

        if self.port_deletion_allowed:
            node_dict["input_ports"] = input_ports
            node_dict["output_ports"] = output_ports

        if self.subgraph_session:
            node_dict["subgraph_session"] = self.subgraph_session

        custom_props = node_dict.pop("_custom_prop", {})
        if custom_props:
            node_dict["custom"] = custom_props

        exclude = [
            "_graph_model",
            "_TEMP_property_attrs",
            "_TEMP_property_widget_types",
        ]
        [node_dict.pop(i) for i in exclude if i in node_dict.keys()]

        return {node_id: node_dict}

    @property
    def serial(self):
        """
        Serialize model information to a string.

        Returns:
            str: serialized JSON string.
        """
        model_dict = self.to_dict
        return json.dumps(model_dict)


class NodeGraphModel(object):
    """
    Data dump for a node graph.
    """

    def __init__(self):
        self.__common_node_props = {}

        self.accept_connection_types = {}
        self.reject_connection_types = {}

        self.nodes = {}
        self.session = ""
        self.acyclic = True
        self.pipe_collision = False
        self.pipe_slicing = True
        self.pipe_style = PipeLayoutEnum.CURVED.value
        self.layout_direction = LayoutDirectionEnum.HORIZONTAL.value

    def common_properties(self):
        """
        Return all common node properties.

        Returns:
            dict: common node properties.
                eg.
                    {'nodeGraphQt.nodes.FooNode': {
                        'my_property': {
                            'widget_type': 0,
                            'tab': 'Properties',
                            'items': ['foo', 'bar', 'test'],
                            'range': (0, 100)
                            }
                        }
                    }
        """
        return self.__common_node_props

    def set_node_common_properties(self, attrs):
        """
        Store common node properties.

        Args:
            attrs (dict): common node properties.
                eg.
                    {'nodeGraphQt.nodes.FooNode': {
                        'my_property': {
                            'widget_type': 0,
                            'tab': 'Properties',
                            'items': ['foo', 'bar', 'test'],
                            'range': (0, 100)
                            }
                        }
                    }
        """
        for node_type in attrs.keys():
            node_props = attrs[node_type]

            if node_type not in self.__common_node_props.keys():
                self.__common_node_props[node_type] = node_props
                continue

            for prop_name, prop_attrs in node_props.items():
                common_props = self.__common_node_props[node_type]
                if prop_name not in common_props.keys():
                    common_props[prop_name] = prop_attrs
                    continue
                common_props[prop_name].update(prop_attrs)

    def get_node_common_properties(self, node_type):
        """
        Return all the common properties for a registered node.

        Args:
            node_type (str): node type.

        Returns:
            dict: node common properties.
        """
        return self.__common_node_props.get(node_type)

    def add_port_accept_connection_type(
        self, port_name, port_type, node_type, accept_pname, accept_ptype, accept_ntype
    ):
        """
        Convenience function for adding to the "accept_connection_types" dict.

        Args:
            port_name (str): current port name.
            port_type (str): current port type.
            node_type (str): current port node type.
            accept_pname (str):port name to accept.
            accept_ptype (str): port type accept.
            accept_ntype (str):port node type to accept.
        """
        connection_data = self.accept_connection_types
        keys = [node_type, port_type, port_name, accept_ntype]
        for key in keys:
            if key not in connection_data.keys():
                connection_data[key] = {}
            connection_data = connection_data[key]

        if accept_ptype not in connection_data:
            connection_data[accept_ptype] = set([accept_pname])
        else:
            # ensure data remains a set instead of list after json de-serialize
            connection_data[accept_ptype] = set(connection_data[accept_ptype]) | {
                accept_pname
            }

    def port_accept_connection_types(self, node_type, port_type, port_name):
        """
        Convenience function for getting the accepted port types from the
        "accept_connection_types" dict.

        Args:
            node_type (str):
            port_type (str):
            port_name (str):

        Returns:
            dict: {<node_type>: {<port_type>: [<port_name>]}}
        """
        data = self.accept_connection_types.get(node_type) or {}
        accepted_types = data.get(port_type) or {}
        return accepted_types.get(port_name) or {}

    def add_port_reject_connection_type(
        self, port_name, port_type, node_type, reject_pname, reject_ptype, reject_ntype
    ):
        """
        Convenience function for adding to the "reject_connection_types" dict.

        Args:
            port_name (str): current port name.
            port_type (str): current port type.
            node_type (str): current port node type.
            reject_pname (str): port name to reject.
            reject_ptype (str): port type to reject.
            reject_ntype (str): port node type to reject.
        """
        connection_data = self.reject_connection_types
        keys = [node_type, port_type, port_name, reject_ntype]
        for key in keys:
            if key not in connection_data.keys():
                connection_data[key] = {}
            connection_data = connection_data[key]

        if reject_ptype not in connection_data:
            connection_data[reject_ptype] = set([reject_pname])
        else:
            # ensure data remains a set instead of list after json de-serialize
            connection_data[reject_ptype] = set(connection_data[reject_ptype]) | {
                reject_pname
            }

    def port_reject_connection_types(self, node_type, port_type, port_name):
        """
        Convenience function for getting the accepted port types from the
        "reject_connection_types" dict.

        Args:
            node_type (str):
            port_type (str):
            port_name (str):

        Returns:
            dict: {<node_type>: {<port_type>: [<port_name>]}}
        """
        data = self.reject_connection_types.get(node_type) or {}
        rejected_types = data.get(port_type) or {}
        return rejected_types.get(port_name) or {}


if __name__ == "__main__":
    p = PortModel(None)
    # print(p.to_dict)

    n = NodeModel()
    n.inputs[p.name] = p
    n.add_property("foo", "bar")

    print("-" * 100)
    print("property keys\n")
    print(list(n.properties.keys()))
    print("-" * 100)
    print("to_dict\n")
    for k, v in n.to_dict[n.id].items():
        print(k, v)



from typing import TypeAlias, Protocol

import polars

from caqtus.session import Shot


class ShotImporter[T](Protocol):
    """Protocol for object that can import a value from a shot.

    A shot importer is a callable that takes a shot returns a value of generic type T.
    """

    def __call__(self, shot: Shot) -> T:
        raise NotImplementedError()


#: Type alias for a data importer that imports data from a shot.
DataImporter: TypeAlias = ShotImporter[polars.DataFrame]



from ._analog_time_lane_exension import analog_time_lane_extension
from ._digital_time_lane_exension import digital_time_lane_extension
from ._camera_time_lane_exension import camera_time_lane_extension
from ._extension import TimeLaneExtension

__all__ = [
    "TimeLaneExtension",
    "digital_time_lane_extension",
    "analog_time_lane_extension",
    "camera_time_lane_extension",
]



#!/usr/bin/python
from collections import OrderedDict

from ..base.commands import NodeVisibleCmd, NodeWidgetVisibleCmd
from ..base.node import NodeObject
from ..base.port import Port
from ..constants import NodePropWidgetEnum, PortTypeEnum
from ..errors import PortError, PortRegistrationError, NodeWidgetError
from ..qgraphics.node_base import NodeItem
from ..widgets.node_widgets import (
    NodeBaseWidget,
    NodeCheckBox,
    NodeComboBox,
    NodeLineEdit,
)


class BaseNode(NodeObject):
    """
    The ``NodeGraphQt.BaseNode`` class is the base class for nodes that allows
    port connections from one node to another.

    .. inheritance-diagram:: NodeGraphQt.BaseNode

    .. image:: ../_images/node.png
        :width: 250px

    example snippet:

    .. code-block:: python
        :linenos:

        from NodeGraphQt import BaseNode

        class ExampleNode(BaseNode):

            # unique node identifier domain.
            __identifier__ = 'io.jchanvfx.github'

            # initial default node name.
            NODE_NAME = 'My Node'

            def __init__(self):
                super(ExampleNode, self).__init__()

                # create an input port.
                self.add_input('in')

                # create an output port.
                self.add_output('out')
    """

    NODE_NAME = "Node"

    def __init__(self, qgraphics_item=None):
        super(BaseNode, self).__init__(qgraphics_item or NodeItem)
        self._inputs = []
        self._outputs = []

    def update_model(self):
        """
        Update the node model from view.
        """
        for name, val in self.view.properties.items():
            if name in ["inputs", "outputs"]:
                continue
            self.model.set_property(name, val)

        for name, widget in self.view.widgets.items():
            self.model.set_property(name, widget.get_value())

    def set_property(self, name, value, push_undo=True):
        """
        Set the value on the node custom property.

        Args:
            name (str): name of the property.
            value (object): property data (python built in types).
            push_undo (bool): register the command to the undo stack. (default: True)
        """
        # prevent signals from causing a infinite loop.
        if self.get_property(name) == value:
            return

        if name == "visible":
            if self.graph:
                undo_cmd = NodeVisibleCmd(self, value)
                if push_undo:
                    self.graph.undo_stack().push(undo_cmd)
                else:
                    undo_cmd.redo()
                return
        elif name == "disabled":
            # redraw the connected pipes in the scene.
            ports = self.view.inputs + self.view.outputs
            for port in ports:
                for pipe in port.connected_pipes:
                    pipe.update()
        super(BaseNode, self).set_property(name, value, push_undo)

    def set_layout_direction(self, value=0):
        """
        Sets the node layout direction to either horizontal or vertical on
        the current node only.

        `Implemented in` ``v0.3.0``

        See Also:
            :meth:`NodeGraph.set_layout_direction`,
            :meth:`NodeObject.layout_direction`


        Warnings:
            This function does not register to the undo stack.

        Args:
            value (int): layout direction mode.
        """
        # base logic to update the model and view attributes only.
        super(BaseNode, self).set_layout_direction(value)
        # redraw the node.
        self._view.draw_node()

    def set_icon(self, icon=None):
        """
        Set the node icon.

        Args:
            icon (str): path to the icon image.
        """
        self.set_property("icon", icon)

    def icon(self):
        """
        Node icon path.

        Returns:
            str: icon image file path.
        """
        return self.model.icon

    def widgets(self):
        """
        Returns all embedded widgets from this node.

        See Also:
            :meth:`BaseNode.get_widget`

        Returns:
            dict: embedded node widgets. {``property_name``: ``node_widget``}
        """
        return self.view.widgets

    def get_widget(self, name):
        """
        Returns the embedded widget associated with the property name.

        See Also:
            :meth:`BaseNode.add_combo_menu`,
            :meth:`BaseNode.add_text_input`,
            :meth:`BaseNode.add_checkbox`,

        Args:
            name (str): node property name.

        Returns:
            NodeBaseWidget: embedded node widget.
        """
        return self.view.widgets.get(name)

    def add_custom_widget(self, widget, widget_type=None, tab=None):
        """
        Add a custom node widget into the node.

        see example :ref:`Embedding Custom Widgets`.

        Note:
            The ``value_changed`` signal from the added node widget is wired
            up to the :meth:`NodeObject.set_property` function.

        Args:
            widget (NodeBaseWidget): node widget class object.
            widget_type: widget flag to display in the
                :class:`NodeGraphQt.PropertiesBinWidget`
                (default: :attr:`NodeGraphQt.constants.NodePropWidgetEnum.HIDDEN`).
            tab (str): name of the widget tab to display in.
        """
        if not isinstance(widget, NodeBaseWidget):
            raise NodeWidgetError("'widget' must be an instance of a NodeBaseWidget")

        widget_type = widget_type or NodePropWidgetEnum.HIDDEN.value
        self.create_property(
            widget.get_name(), widget.get_value(), widget_type=widget_type, tab=tab
        )
        widget.value_changed.connect(lambda k, v: self.set_property(k, v))
        widget._node = self
        self.view.add_widget(widget)
        #: redraw node to address calls outside the "__init__" func.
        self.view.draw_node()

        #: HACK: calling the .parent() function here on the widget as it seems
        #        to address a seg fault issue when exiting the application.
        widget.parent()

    def add_combo_menu(self, name, label="", items=None, tooltip=None, tab=None):
        """
        Creates a custom property with the :meth:`NodeObject.create_property`
        function and embeds a :class:`PySide2.QtWidgets.QComboBox` widget
        into the node.

        Note:
            The ``value_changed`` signal from the added node widget is wired
            up to the :meth:`NodeObject.set_property` function.

        Args:
            name (str): name for the custom property.
            label (str): label to be displayed.
            items (list[str]): items to be added into the menu.
            tooltip (str): widget tooltip.
            tab (str): name of the widget tab to display in.
        """
        self.create_property(
            name,
            value=items[0] if items else None,
            items=items or [],
            widget_type=NodePropWidgetEnum.QCOMBO_BOX.value,
            widget_tooltip=tooltip,
            tab=tab,
        )
        widget = NodeComboBox(self.view, name, label, items)
        widget.setToolTip(tooltip or "")
        widget.value_changed.connect(lambda k, v: self.set_property(k, v))
        self.view.add_widget(widget)
        #: redraw node to address calls outside the "__init__" func.
        self.view.draw_node()

    def add_text_input(
        self, name, label="", text="", placeholder_text="", tooltip=None, tab=None
    ):
        """
        Creates a custom property with the :meth:`NodeObject.create_property`
        function and embeds a :class:`PySide2.QtWidgets.QLineEdit` widget
        into the node.

        Note:
            The ``value_changed`` signal from the added node widget is wired
            up to the :meth:`NodeObject.set_property` function.

        Args:
            name (str): name for the custom property.
            label (str): label to be displayed.
            text (str): pre-filled text.
            placeholder_text (str): placeholder text.
            tooltip (str): widget tooltip.
            tab (str): name of the widget tab to display in.
        """
        self.create_property(
            name,
            value=text,
            widget_type=NodePropWidgetEnum.QLINE_EDIT.value,
            widget_tooltip=tooltip,
            tab=tab,
        )
        widget = NodeLineEdit(self.view, name, label, text, placeholder_text)
        widget.setToolTip(tooltip or "")
        widget.value_changed.connect(lambda k, v: self.set_property(k, v))
        self.view.add_widget(widget)
        #: redraw node to address calls outside the "__init__" func.
        self.view.draw_node()

    def add_checkbox(
        self, name, label="", text="", state=False, tooltip=None, tab=None
    ):
        """
        Creates a custom property with the :meth:`NodeObject.create_property`
        function and embeds a :class:`PySide2.QtWidgets.QCheckBox` widget
        into the node.

        Note:
            The ``value_changed`` signal from the added node widget is wired
            up to the :meth:`NodeObject.set_property` function.

        Args:
            name (str): name for the custom property.
            label (str): label to be displayed.
            text (str): checkbox text.
            state (bool): pre-check.
            tooltip (str): widget tooltip.
            tab (str): name of the widget tab to display in.
        """
        self.create_property(
            name,
            value=state,
            widget_type=NodePropWidgetEnum.QCHECK_BOX.value,
            widget_tooltip=tooltip,
            tab=tab,
        )
        widget = NodeCheckBox(self.view, name, label, text, state)
        widget.setToolTip(tooltip or "")
        widget.value_changed.connect(lambda k, v: self.set_property(k, v))
        self.view.add_widget(widget)
        #: redraw node to address calls outside the "__init__" func.
        self.view.draw_node()

    def hide_widget(self, name, push_undo=True):
        """
        Hide an embedded node widget.

        Args:
            name (str): node property name for the widget.
            push_undo (bool): register the command to the undo stack. (default: True)

        See Also:
            :meth:`BaseNode.add_custom_widget`,
            :meth:`BaseNode.show_widget`,
            :meth:`BaseNode.get_widget`
        """
        if not self.view.has_widget(name):
            return
        undo_cmd = NodeWidgetVisibleCmd(self, name, visible=False)
        if push_undo:
            self.graph.undo_stack().push(undo_cmd)
        else:
            undo_cmd.redo()

    def show_widget(self, name, push_undo=True):
        """
        Show an embedded node widget.

        Args:
            name (str): node property name for the widget.
            push_undo (bool): register the command to the undo stack. (default: True)

        See Also:
            :meth:`BaseNode.add_custom_widget`,
            :meth:`BaseNode.hide_widget`,
            :meth:`BaseNode.get_widget`
        """
        if not self.view.has_widget(name):
            return
        undo_cmd = NodeWidgetVisibleCmd(self, name, visible=True)
        if push_undo:
            self.graph.undo_stack().push(undo_cmd)
        else:
            undo_cmd.redo()

    def add_input(
        self,
        name="input",
        multi_input=False,
        display_name=True,
        color=None,
        locked=False,
        painter_func=None,
    ):
        """
        Add input :class:`Port` to node.

        Warnings:
            Undo is NOT supported for this function.

        Args:
            name (str): name for the input port.
            multi_input (bool): allow port to have more than one connection.
            display_name (bool): display the port name on the node.
            color (tuple): initial port color (r, g, b) ``0-255``.
            locked (bool): locked state see :meth:`Port.set_locked`
            painter_func (function or None): custom function to override the drawing
                of the port shape see example: :ref:`Creating Custom Shapes`

        Returns:
            NodeGraphQt.Port: the created port object.
        """
        if name in self.inputs().keys():
            raise PortRegistrationError(
                'port name "{}" already registered.'.format(name)
            )

        port_args = [name, multi_input, display_name, locked]
        if painter_func and callable(painter_func):
            port_args.append(painter_func)
        view = self.view.add_input(*port_args)

        if color:
            view.color = color
            view.border_color = [min([255, max([0, i + 80])]) for i in color]

        port = Port(self, view)
        port.model.type_ = PortTypeEnum.IN.value
        port.model.name = name
        port.model.display_name = display_name
        port.model.multi_connection = multi_input
        port.model.locked = locked
        self._inputs.append(port)
        self.model.inputs[port.name()] = port.model
        return port

    def add_output(
        self,
        name="output",
        multi_output=True,
        display_name=True,
        color=None,
        locked=False,
        painter_func=None,
    ):
        """
        Add output :class:`Port` to node.

        Warnings:
            Undo is NOT supported for this function.

        Args:
            name (str): name for the output port.
            multi_output (bool): allow port to have more than one connection.
            display_name (bool): display the port name on the node.
            color (tuple): initial port color (r, g, b) ``0-255``.
            locked (bool): locked state see :meth:`Port.set_locked`
            painter_func (function or None): custom function to override the drawing
                of the port shape see example: :ref:`Creating Custom Shapes`

        Returns:
            NodeGraphQt.Port: the created port object.
        """
        if name in self.outputs().keys():
            raise PortRegistrationError(
                'port name "{}" already registered.'.format(name)
            )

        port_args = [name, multi_output, display_name, locked]
        if painter_func and callable(painter_func):
            port_args.append(painter_func)
        view = self.view.add_output(*port_args)

        if color:
            view.color = color
            view.border_color = [min([255, max([0, i + 80])]) for i in color]
        port = Port(self, view)
        port.model.type_ = PortTypeEnum.OUT.value
        port.model.name = name
        port.model.display_name = display_name
        port.model.multi_connection = multi_output
        port.model.locked = locked
        self._outputs.append(port)
        self.model.outputs[port.name()] = port.model
        return port

    def get_input(self, port):
        """
        Get input port by the name or index.

        Args:
            port (str or int): port name or index.

        Returns:
            NodeGraphQt.Port: node port.
        """
        if type(port) is int:
            if port < len(self._inputs):
                return self._inputs[port]
        elif type(port) is str:
            return self.inputs().get(port, None)

    def get_output(self, port):
        """
        Get output port by the name or index.

        Args:
            port (str or int): port name or index.

        Returns:
            NodeGraphQt.Port: node port.
        """
        if type(port) is int:
            if port < len(self._outputs):
                return self._outputs[port]
        elif type(port) is str:
            return self.outputs().get(port, None)

    def delete_input(self, port):
        """
        Delete input port.

        Warnings:
            Undo is NOT supported for this function.

            You can only delete ports if :meth:`BaseNode.port_deletion_allowed`
            returns ``True`` otherwise a port error is raised see also
            :meth:`BaseNode.set_port_deletion_allowed`.

        Args:
            port (str or int): port name or index.
        """
        if type(port) in [int, str]:
            port = self.get_input(port)
            if port is None:
                return
        if not self.port_deletion_allowed():
            raise PortError(
                'Port "{}" can\'t be deleted on this node because '
                '"ports_removable" is not enabled.'.format(port.name())
            )
        if port.locked():
            raise PortError("Error: Can't delete a port that is locked!")
        self._inputs.remove(port)
        self._model.inputs.pop(port.name())
        self._view.delete_input(port.view)
        port.model.node = None
        self._view.draw_node()

    def delete_output(self, port):
        """
        Delete output port.

        Warnings:
            Undo is NOT supported for this function.

            You can only delete ports if :meth:`BaseNode.port_deletion_allowed`
            returns ``True`` otherwise a port error is raised see also
            :meth:`BaseNode.set_port_deletion_allowed`.

        Args:
            port (str or int): port name or index.
        """
        if type(port) in [int, str]:
            port = self.get_output(port)
            if port is None:
                return
        if not self.port_deletion_allowed():
            raise PortError(
                'Port "{}" can\'t be deleted on this node because '
                '"ports_removable" is not enabled.'.format(port.name())
            )
        if port.locked():
            raise PortError("Error: Can't delete a port that is locked!")
        self._outputs.remove(port)
        self._model.outputs.pop(port.name())
        self._view.delete_output(port.view)
        port.model.node = None
        self._view.draw_node()

    def set_port_deletion_allowed(self, mode=False):
        """
        Allow ports to be removable on this node.

        See Also:
            :meth:`BaseNode.port_deletion_allowed` and
            :meth:`BaseNode.set_ports`

        Args:
            mode (bool): true to allow.
        """
        self.model.port_deletion_allowed = mode

    def port_deletion_allowed(self):
        """
        Return true if ports can be deleted on this node.

        See Also:
            :meth:`BaseNode.set_port_deletion_allowed`

        Returns:
            bool: true if ports can be deleted.
        """
        return self.model.port_deletion_allowed

    def set_ports(self, port_data):
        """
        Create node input and output ports from serialized port data.

        Warnings:
            You can only use this function if the node has
            :meth:`BaseNode.port_deletion_allowed` is `True`
            see :meth:`BaseNode.set_port_deletion_allowed`

        Hint:
            example snippet of port data.

            .. highlight:: python
            .. code-block:: python

                {
                    'input_ports':
                        [{
                            'name': 'input',
                            'multi_connection': True,
                            'display_name': 'Input',
                            'locked': False
                        }],
                    'output_ports':
                        [{
                            'name': 'output',
                            'multi_connection': True,
                            'display_name': 'Output',
                            'locked': False
                        }]
                }

        Args:
            port_data(dict): port data.
        """
        if not self.port_deletion_allowed():
            raise PortError(
                "Ports cannot be set on this node because "
                '"set_port_deletion_allowed" is not enabled on this node.'
            )

        for port in self._inputs:
            self._view.delete_input(port.view)
            port.model.node = None
        for port in self._outputs:
            self._view.delete_output(port.view)
            port.model.node = None
        self._inputs = []
        self._outputs = []
        self._model.outputs = {}
        self._model.inputs = {}

        [
            self.add_input(
                name=port["name"],
                multi_input=port["multi_connection"],
                display_name=port["display_name"],
                locked=port.get("locked") or False,
            )
            for port in port_data["input_ports"]
        ]
        [
            self.add_output(
                name=port["name"],
                multi_output=port["multi_connection"],
                display_name=port["display_name"],
                locked=port.get("locked") or False,
            )
            for port in port_data["output_ports"]
        ]
        self._view.draw_node()

    def inputs(self):
        """
        Returns all the input ports from the node.

        Returns:
            dict: {<port_name>: <port_object>}
        """
        return {p.name(): p for p in self._inputs}

    def input_ports(self):
        """
        Return all input ports.

        Returns:
            list[NodeGraphQt.Port]: node input ports.
        """
        return self._inputs

    def outputs(self):
        """
        Returns all the output ports from the node.

        Returns:
            dict: {<port_name>: <port_object>}
        """
        return {p.name(): p for p in self._outputs}

    def output_ports(self):
        """
        Return all output ports.

        Returns:
            list[NodeGraphQt.Port]: node output ports.
        """
        return self._outputs

    def input(self, index):
        """
        Return the input port with the matching index.

        Args:
            index (int): index of the input port.

        Returns:
            NodeGraphQt.Port: port object.
        """
        return self._inputs[index]

    def set_input(self, index, port):
        """
        Creates a connection pipe to the targeted output :class:`Port`.

        Args:
            index (int): index of the port.
            port (NodeGraphQt.Port): port object.
        """
        src_port = self.input(index)
        src_port.connect_to(port)

    def output(self, index):
        """
        Return the output port with the matching index.

        Args:
            index (int): index of the output port.

        Returns:
            NodeGraphQt.Port: port object.
        """
        return self._outputs[index]

    def set_output(self, index, port):
        """
        Creates a connection pipe to the targeted input :class:`Port`.

        Args:
            index (int): index of the port.
            port (NodeGraphQt.Port): port object.
        """
        src_port = self.output(index)
        src_port.connect_to(port)

    def connected_input_nodes(self):
        """
        Returns all nodes connected from the input ports.

        Returns:
            dict: {<input_port>: <node_list>}
        """
        nodes = OrderedDict()
        for p in self.input_ports():
            nodes[p] = [cp.node() for cp in p.connected_ports()]
        return nodes

    def connected_output_nodes(self):
        """
        Returns all nodes connected from the output ports.

        Returns:
            dict: {<output_port>: <node_list>}
        """
        nodes = OrderedDict()
        for p in self.output_ports():
            nodes[p] = [cp.node() for cp in p.connected_ports()]
        return nodes

    def add_accept_port_type(self, port, port_type_data):
        """
        Add an accept constrain to a specified node port.

        Once a constraint has been added only ports of that type specified will
        be allowed a pipe connection.

        port type data example

        .. highlight:: python
        .. code-block:: python

            {
                'port_name': 'foo'
                'port_type': PortTypeEnum.IN.value
                'node_type': 'io.github.jchanvfx.NodeClass'
            }

        See Also:
            :meth:`NodeGraphQt.BaseNode.accepted_port_types`

        Args:
            port (NodeGraphQt.Port): port to assign constrain to.
            port_type_data (dict): port type data to accept a connection
        """
        node_ports = self._inputs + self._outputs
        if port not in node_ports:
            raise PortError('Node does not contain port: "{}"'.format(port))

        self._model.add_port_accept_connection_type(
            port_name=port.name(),
            port_type=port.type_(),
            node_type=self.type_,
            accept_pname=port_type_data["port_name"],
            accept_ptype=port_type_data["port_type"],
            accept_ntype=port_type_data["node_type"],
        )

    def accepted_port_types(self, port):
        """
        Returns a dictionary of connection constrains of the port types
        that allow for a pipe connection to this node.

        Args:
            port (NodeGraphQt.Port): port object.

        Returns:
            dict: {<node_type>: {<port_type>: [<port_name>]}}
        """
        ports = self._inputs + self._outputs
        if port not in ports:
            raise PortError('Node does not contain port "{}"'.format(port))

        accepted_types = self.graph.model.port_accept_connection_types(
            node_type=self.type_, port_type=port.type_(), port_name=port.name()
        )
        return accepted_types

    def add_reject_port_type(self, port, port_type_data):
        """
        Add a reject constrain to a specified node port.

        Once a constraint has been added only ports of that type specified will
        NOT be allowed a pipe connection.

        port type data example

        .. highlight:: python
        .. code-block:: python

            {
                'port_name': 'foo'
                'port_type': PortTypeEnum.IN.value
                'node_type': 'io.github.jchanvfx.NodeClass'
            }

        See Also:
            :meth:`NodeGraphQt.Port.rejected_port_types`

        Args:
            port (NodeGraphQt.Port): port to assign constrain to.
            port_type_data (dict): port type data to reject a connection
        """
        node_ports = self._inputs + self._outputs
        if port not in node_ports:
            raise PortError('Node does not contain port: "{}"'.format(port))

        self._model.add_port_reject_connection_type(
            port_name=port.name(),
            port_type=port.type_(),
            node_type=self.type_,
            reject_pname=port_type_data["port_name"],
            reject_ptype=port_type_data["port_type"],
            reject_ntype=port_type_data["node_type"],
        )

    def rejected_port_types(self, port):
        """
        Returns a dictionary of connection constrains of the port types
        that are NOT allowed for a pipe connection to this node.

        Args:
            port (NodeGraphQt.Port): port object.

        Returns:
            dict: {<node_type>: {<port_type>: [<port_name>]}}
        """
        ports = self._inputs + self._outputs
        if port not in ports:
            raise PortError('Node does not contain port "{}"'.format(port))

        rejected_types = self.graph.model.port_reject_connection_types(
            node_type=self.type_, port_type=port.type_(), port_name=port.name()
        )
        return rejected_types

    def on_input_connected(self, in_port, out_port):
        """
        Callback triggered when a new pipe connection is made.

        *The default of this function does nothing re-implement if you require
        logic to run for this event.*

        Note:
            to work with undo & redo for this method re-implement
            :meth:`BaseNode.on_input_disconnected` with the reverse logic.

        Args:
            in_port (NodeGraphQt.Port): source input port from this node.
            out_port (NodeGraphQt.Port): output port that connected to this node.
        """
        return

    def on_input_disconnected(self, in_port, out_port):
        """
        Callback triggered when a pipe connection has been disconnected
        from a INPUT port.

        *The default of this function does nothing re-implement if you require
        logic to run for this event.*

        Note:
            to work with undo & redo for this method re-implement
            :meth:`BaseNode.on_input_connected` with the reverse logic.

        Args:
            in_port (NodeGraphQt.Port): source input port from this node.
            out_port (NodeGraphQt.Port): output port that was disconnected.
        """
        return



from collections.abc import Mapping
from typing import Any

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.shot_compilation import DeviceCompiler
from caqtus.shot_compilation.compilation_contexts import ShotContext, SequenceContext
from caqtus.shot_compilation.variable_namespace import VariableNamespace
from caqtus.types.recoverable_exceptions import InvalidValueError
from caqtus.types.timelane import TimeLanes


class ShotCompiler:
    def __init__(
        self,
        shot_timelanes: TimeLanes,
        device_configurations: Mapping[DeviceName, DeviceConfiguration],
        device_compilers: Mapping[DeviceName, DeviceCompiler],
    ):
        self.shot_time_lanes = shot_timelanes
        self.device_configurations = device_configurations
        self._sequence_context = SequenceContext(
            device_configurations=device_configurations, time_lanes=shot_timelanes
        )
        self.device_compilers = device_compilers

    def compile_shot(
        self, shot_parameters: VariableNamespace
    ) -> tuple[Mapping[DeviceName, Mapping[str, Any]], float]:
        shot_context = ShotContext(
            sequence_context=self._sequence_context,
            variables=shot_parameters.dict(),
            device_compilers=self.device_compilers,
        )

        results = {}
        for device_name, compiler in self.device_compilers.items():
            results[device_name] = shot_context.get_shot_parameters(device_name)

        # noinspection PyProtectedMember
        if unused_lanes := shot_context._unused_lanes():
            raise InvalidValueError(
                "The following lanes where not used during the shot: "
                + ", ".join(unused_lanes)
            )

        return results, shot_context.get_shot_duration()



from caqtus.extension import Experiment
from caqtus.session.sql import PostgreSQLConfig

exp = Experiment()

exp.configure_storage(PostgreSQLConfig.from_file("config.yaml"))

exp.setup_default_extensions()

if __name__ == "__main__":

    exp.launch_condetrol()



from ._async_converter import AsyncConverter
from ._device_proxy import DeviceProxy
from .rpc import (
    Server,
    RPCConfiguration,
    InsecureRPCConfiguration,
)
from ._proxy import Proxy

__all__ = [
    "DeviceProxy",
    "Server",
    "RPCConfiguration",
    "InsecureRPCConfiguration",
    "AsyncConverter",
    "Proxy",
]



from collections.abc import Callable
from typing import Concatenate, ParamSpec, TypeVar

import attrs

from caqtus.device.configuration import DeviceServerName
from caqtus.device.configuration.serializer import DeviceConfigJSONSerializer
from caqtus.device.remote import RPCConfiguration
from caqtus.experiment_control.device_manager_extension import DeviceManagerExtension
from caqtus.gui.condetrol._extension import CondetrolExtension
from caqtus.session import ExperimentSessionMaker
from caqtus.session.sql._serializer import SerializerProtocol, Serializer
from caqtus.types.timelane._serializer import TimeLaneSerializer
from ._protocol import CaqtusExtensionProtocol
from ..device_extension import DeviceExtension
from ..time_lane_extension import TimeLaneExtension

P = ParamSpec("P")
T = TypeVar("T", bound=ExperimentSessionMaker)


@attrs.frozen
class CaqtusExtension(CaqtusExtensionProtocol):
    condetrol_extension: CondetrolExtension = attrs.field(factory=CondetrolExtension)
    device_configurations_serializer: DeviceConfigJSONSerializer = attrs.field(
        factory=DeviceConfigJSONSerializer
    )
    time_lane_serializer: TimeLaneSerializer = attrs.field(factory=TimeLaneSerializer)
    device_manager_extension: DeviceManagerExtension = attrs.field(
        factory=DeviceManagerExtension
    )

    def __attrs_post_init__(self):
        self.condetrol_extension.lane_extension.set_lane_serializer(
            self.time_lane_serializer
        )

    def register_device_extension(self, device_extension: DeviceExtension) -> None:
        self.condetrol_extension.device_extension.register_device_configuration_editor(
            device_extension.configuration_type, device_extension.editor_type
        )
        self.condetrol_extension.device_extension.register_configuration_factory(
            device_extension.label, device_extension.configuration_factory
        )
        self.device_configurations_serializer.register_device_configuration(
            device_extension.configuration_type,
            device_extension.configuration_dumper,
            device_extension.configuration_loader,
        )
        self.device_manager_extension.register_device_compiler(
            device_extension.configuration_type, device_extension.compiler_type
        )
        self.device_manager_extension.register_controller(
            device_extension.configuration_type, device_extension.controller_type
        )
        self.device_manager_extension.register_device(
            device_extension.configuration_type, device_extension.device_type
        )
        self.device_manager_extension.register_proxy(
            device_extension.configuration_type, device_extension.proxy_type
        )

    def register_time_lane_extension(
        self, time_lane_extension: TimeLaneExtension
    ) -> None:
        self.time_lane_serializer.register_time_lane(
            time_lane_extension.lane_type,
            time_lane_extension.dumper,
            time_lane_extension.loader,
            time_lane_extension.type_tag,
        )
        self.condetrol_extension.lane_extension.register_lane_factory(
            time_lane_extension.label, time_lane_extension.lane_factory
        )
        self.condetrol_extension.lane_extension.register_lane_model_factory(
            time_lane_extension.lane_type, time_lane_extension.lane_model_factory
        )
        self.condetrol_extension.lane_extension.register_lane_delegate_factory(
            time_lane_extension.lane_type, time_lane_extension.lane_delegate_factory
        )

    def register_device_server_config(
        self,
        name: DeviceServerName,
        config: RPCConfiguration,
    ) -> None:
        self.device_manager_extension.register_device_server_config(name, config)

    def create_session_maker(
        self,
        session_maker_type: Callable[Concatenate[SerializerProtocol, P], T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        serializer = Serializer.default()
        serializer.device_configuration_serializer = (
            self.device_configurations_serializer
        )
        serializer.time_lane_serializer = self.time_lane_serializer
        return session_maker_type(serializer, *args, **kwargs)



from __future__ import annotations

from numbers import Integral, Real
from typing import Mapping, Any, Optional

import attrs

import caqtus.formatter as fmt
from caqtus.device.sequencer.channel_commands.channel_output import (
    ChannelOutput,
    DimensionedSeries,
)
from caqtus.device.sequencer.instructions import Pattern
from caqtus.shot_compilation import ShotContext
from caqtus.shot_compilation.lane_compilers.timing import number_ticks, ns
from caqtus.types.expression import Expression
from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.types.units import Quantity, Unit, dimensionless
from caqtus.types.variable_name import DottedVariableName
from ..._time_step import TimeStep


@attrs.define
class Constant(ChannelOutput):
    """Indicates that the output should be held at a constant value during the shot.

    The constant value is obtained by evaluating the value stored in the constant
    output within the shot context.
    Note that `constant` refers to a value constant in shot time, not necessarily
    constant across shots.
    """

    value: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return str(self.value)

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ) -> DimensionedSeries:
        length = (
            prepend
            + number_ticks(0, shot_context.get_shot_duration(), required_time_step * ns)
            + append
        )
        value = self.value.evaluate(shot_context.get_variables())
        magnitude, units = split_magnitude_units(value)
        return DimensionedSeries(Pattern([magnitude]) * length, units)

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        return 0, 0


def split_magnitude_units(value: Any) -> tuple[bool | int | float, Optional[Unit]]:
    if isinstance(value, Quantity):
        in_base_units = value.to_base_units()
        magnitude = float(in_base_units.magnitude)
        units = in_base_units.units if in_base_units.units != dimensionless else None
    elif isinstance(value, bool):
        magnitude = value
        units = None
    elif isinstance(value, Integral):
        magnitude = int(value)
        units = None
    elif isinstance(value, Real):
        magnitude = float(value)
        units = None
    else:
        raise InvalidTypeError(
            f"Constant value must be a number or a boolean, got {fmt.type_(value)}"
        )
    assert isinstance(units, Unit) or units is None
    return magnitude, units



# Form implementation generated from reading ui file '.\graphplot\errorbar_visualizer.ui'
#
# Created by: PyQt6 UI code generator 6.6.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtWidgets


class Ui_ErrorBarVisualizerCreator(object):
    def setupUi(self, ErrorBarVisualizerCreator):
        ErrorBarVisualizerCreator.setObjectName("ErrorBarVisualizerCreator")
        ErrorBarVisualizerCreator.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(ErrorBarVisualizerCreator)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.xAxisLabel = QtWidgets.QLabel(parent=ErrorBarVisualizerCreator)
        self.xAxisLabel.setObjectName("xAxisLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.xAxisLabel)
        self._x_axis_line_edit = QtWidgets.QLineEdit(parent=ErrorBarVisualizerCreator)
        self._x_axis_line_edit.setObjectName("_x_axis_line_edit")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self._x_axis_line_edit)
        self.yAxisLabel = QtWidgets.QLabel(parent=ErrorBarVisualizerCreator)
        self.yAxisLabel.setObjectName("yAxisLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.LabelRole, self.yAxisLabel)
        self._y_axis_line_edit = QtWidgets.QLineEdit(parent=ErrorBarVisualizerCreator)
        self._y_axis_line_edit.setObjectName("_y_axis_line_edit")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.FieldRole, self._y_axis_line_edit)
        self.hueLabel = QtWidgets.QLabel(parent=ErrorBarVisualizerCreator)
        self.hueLabel.setObjectName("hueLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.hueLabel)
        self._hue_line_edit = QtWidgets.QLineEdit(parent=ErrorBarVisualizerCreator)
        self._hue_line_edit.setObjectName("_hue_line_edit")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.FieldRole, self._hue_line_edit)
        self.verticalLayout.addLayout(self.formLayout)

        self.retranslateUi(ErrorBarVisualizerCreator)
        QtCore.QMetaObject.connectSlotsByName(ErrorBarVisualizerCreator)

    def retranslateUi(self, ErrorBarVisualizerCreator):
        _translate = QtCore.QCoreApplication.translate
        ErrorBarVisualizerCreator.setWindowTitle(_translate("ErrorBarVisualizerCreator", "Form"))
        self.xAxisLabel.setText(_translate("ErrorBarVisualizerCreator", "x-axis"))
        self.yAxisLabel.setText(_translate("ErrorBarVisualizerCreator", "y-axis"))
        self.hueLabel.setText(_translate("ErrorBarVisualizerCreator", "hue"))
        self._hue_line_edit.setPlaceholderText(_translate("ErrorBarVisualizerCreator", "None"))



"""Allow to manage the high-level control of experiments."""

from . import shot_timing
from .manager import ExperimentManager, Procedure
from .sequence_runner import ShotRetryConfig

__all__ = [
    "ShotRetryConfig",
    "ExperimentManager",
    "Procedure",
    "shot_timing",
]



import heapq
import math

import multipledispatch
import numpy as np

from caqtus.utils.itertools import pairwise
from ._instructions import (
    SequencerInstruction,
    Pattern,
    Concatenated,
    Repeated,
    empty_like,
    concatenate,
)
from ._with_name import with_name


def merge_instructions(**instructions: SequencerInstruction) -> SequencerInstruction:
    """Merge several instructions by name.

    This function finds a common structure to the different instructions and produces
    a single instruction with parallel fields for each input instruction.

    Args:
        instructions: The instructions to merge by name.
            There must be at least one instruction.
            They must all have the same length.

    Returns:
        A new instruction with the same length as the input instructions,
        and a structured dtype with a field for each input instruction.

    Warning:
        If the input instructions have no simple common structure, this function will
        convert each instruction to an explicit pattern and merge them.
        If the input instructions have a very long length, this function might be slow
        and consume a lot of memory.

    Raises:
        ValueError: If the instructions have different lengths or no instructions are
            provided.
    """

    if not instructions:
        raise ValueError("No instructions to merge")

    # Check that all instructions have the same length
    length = len(next(iter(instructions.values())))
    for instruction in instructions.values():
        if len(instruction) != length:
            raise ValueError("Instructions must have the same length")

    named_instructions = [
        with_name(instruction, name) for name, instruction in instructions.items()
    ]
    return _stack_instructions_no_checks(*named_instructions)


def stack_instructions(
    *instructions: SequencerInstruction[np.void],
) -> SequencerInstruction[np.void]:
    """Stack several instructions along their dtype names.

    Args:
        instructions: A sequence of instructions to stack.
            They must all have the same length.
            They must have a structured dtype with named fields.

    Returns:
        A new instruction with the same length as the input instructions,
        and a dtype that is the union of the input dtypes.
    """

    if not instructions:
        raise ValueError("No instructions to stack")

    # Check that all instructions have the same length
    length = len(instructions[0])
    for instruction in instructions[1:]:
        if len(instruction) != length:
            raise ValueError("Instructions must have the same length")

    # Check that all instructions have a structured dtype
    for instruction in instructions:
        if not issubclass(instruction.dtype.type, np.void):
            raise ValueError("Instruction must have a structured dtype")

    return _stack_instructions_no_checks(*instructions)


def _stack_instructions_no_checks(
    *instructions: SequencerInstruction[np.void],
) -> SequencerInstruction:
    # This uses a divide-and-conquer approach to merge the instructions.
    # Another approach is to stack the instructions into a single accumulator, but
    # it seems to give worse performance on typical uses.

    if len(instructions) == 1:
        return instructions[0]
    elif len(instructions) == 2:
        return stack(instructions[0], instructions[1])
    else:
        length = len(instructions) // 2
        sub_block_1 = _stack_instructions_no_checks(*instructions[:length])
        sub_block_2 = _stack_instructions_no_checks(*instructions[length:])
        return stack(sub_block_1, sub_block_2)


stack = multipledispatch.Dispatcher("stack")


@stack.register(SequencerInstruction, SequencerInstruction)
def stack_generic(
    a: SequencerInstruction[np.void], b: SequencerInstruction[np.void]
) -> SequencerInstruction:
    assert len(a) == len(b)
    return _stack_patterns(a.to_pattern(), b.to_pattern())


def _stack_patterns(a: Pattern[np.void], b: Pattern[np.void]) -> Pattern[np.void]:
    merged_dtype = merge_dtypes(a.dtype, b.dtype)
    merged = np.zeros(len(a), dtype=merged_dtype)

    assert a.dtype.names is not None
    for name in a.dtype.names:
        merged[name] = a.array[name]

    assert b.dtype.names is not None
    for name in b.dtype.names:
        merged[name] = b.array[name]
    return Pattern.create_without_copy(merged)


@stack.register(Concatenated, Concatenated)
def stack_concatenations(a: Concatenated, b: Concatenated) -> SequencerInstruction:
    assert len(a) == len(b)

    new_bounds = heapq.merge(a._instruction_bounds, b._instruction_bounds)
    results = []
    for start, stop in pairwise(new_bounds):
        results.append(stack(a[start:stop], b[start:stop]))
    if not results:
        return stack(empty_like(a), empty_like(b))
    return concatenate(*results)


@stack.register(Concatenated, SequencerInstruction)
def stack_concatenation_left(
    a: Concatenated, b: SequencerInstruction
) -> SequencerInstruction:
    assert len(a) == len(b)

    results = []
    for (start, stop), instruction in zip(
        pairwise(a._instruction_bounds), a.instructions
    ):
        results.append(stack(instruction, b[start:stop]))
    if not results:
        return stack(empty_like(a), empty_like(b))
    return concatenate(*results)


@stack.register(SequencerInstruction, Concatenated)
def stack_concatenation_right(
    a: SequencerInstruction, b: Concatenated
) -> SequencerInstruction:
    assert len(a) == len(b)

    results = []
    for (start, stop), instruction in zip(
        pairwise(b._instruction_bounds), b.instructions
    ):
        results.append(stack(a[start:stop], instruction))
    if not results:
        return stack(empty_like(a), empty_like(b))
    return concatenate(*results)


@stack.register(Repeated, Repeated)
def stack_repeated(a: Repeated, b: Repeated) -> SequencerInstruction:
    assert len(a) == len(b)

    lcm = math.lcm(len(a.instruction), len(b.instruction))
    if lcm == len(a):
        b_a = tile(a.instruction, a.repetitions)
        b_b = tile(b.instruction, b.repetitions)
    else:
        r_a = lcm // len(a.instruction)
        b_a = a.instruction * r_a
        r_b = lcm // len(b.instruction)
        b_b = b.instruction * r_b
    block = stack(b_a, b_b)
    return block * (len(a) // len(block))


def merge_dtypes(a: np.dtype[np.void], b: np.dtype[np.void]) -> np.dtype[np.void]:
    assert a.names is not None
    assert b.names is not None
    merged_dtype = np.dtype(
        [(name, a[name]) for name in a.names] + [(name, b[name]) for name in b.names]
    )
    return merged_dtype


def tile[
    T: np.generic
](instruction: SequencerInstruction[T], repetitions: int) -> SequencerInstruction[T]:
    return concatenate(*([instruction] * repetitions))



from __future__ import annotations

from typing import Optional, Any

import polars
from PyQt6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PyQt6.QtWidgets import QTableView
from caqtus.analysis.units import is_quantity_dtype
from caqtus.types.units import Quantity

from ..visualizer_creator import ViewCreator, DataView


class DataFrameViewCreator(ViewCreator):
    def __init__(self) -> None:
        super().__init__()

    def create_view(self) -> DataFrameView:
        return DataFrameView()


class DataFrameView(QTableView, DataView):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._model = DataFrameModel(polars.DataFrame())
        self.setModel(self._model)

    def update_data(self, dataframe: Optional[polars.DataFrame]) -> None:
        if dataframe is None:
            dataframe = polars.DataFrame()
        self._model.update_dataframe(dataframe)


class DataFrameModel(QAbstractTableModel):
    def __init__(self, dataframe: polars.DataFrame, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._dataframe = dataframe

    def update_dataframe(self, dataframe: polars.DataFrame) -> None:
        self.beginResetModel()
        self._dataframe = dataframe
        self.endResetModel()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._dataframe)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._dataframe.columns)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            column_name = self._dataframe.columns[index.column()]
            column = self._dataframe[column_name]
            value = column[index.row()]
            if is_quantity_dtype(column.dtype):
                quantity = Quantity(value["magnitude"], value["units"])
                return format(quantity, "~")
            else:
                return str(value)

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self._dataframe.columns[section]
            elif orientation == Qt.Orientation.Vertical:
                return section



#!/usr/bin/python
from PySide6 import QtWidgets, QtCore


class BaseProperty(QtWidgets.QWidget):
    """
    Base class for a custom node property widget to be displayed in the
    PropertiesBin widget.

    Inherits from: :class:`PySide2.QtWidgets.QWidget`
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(BaseProperty, self).__init__(parent)
        self._name = None

    def __repr__(self):
        return "<{}() object at {}>".format(self.__class__.__name__, hex(id(self)))

    def get_name(self):
        """
        Returns:
            str: property name matching the node property.
        """
        return self._name

    def set_name(self, name):
        """
        Args:
            name (str): property name matching the node property.
        """
        self._name = name

    def get_value(self):
        """
        Returns:
            object: widgets current value.
        """
        raise NotImplementedError

    def set_value(self, value):
        """
        Args:
            value (object): property value to update the widget.
        """
        raise NotImplementedError



"""Package to control and analyze cold atom experiments."""

from . import (
    analysis,
    device,
    experiment_control,
    extension,
    gui,
    session,
    shot_compilation,
    types,
    utils,
)


__all__ = [
    "analysis",
    "device",
    "experiment_control",
    "extension",
    "gui",
    "session",
    "shot_compilation",
    "types",
    "utils",
]



import os

from PySide6 import QtWidgets, QtGui, QtCore

_current_user_directory = os.path.expanduser("~")


def _set_dir(file):
    global _current_user_directory
    if os.path.isdir(file):
        _current_user_directory = file
    elif os.path.isfile(file):
        _current_user_directory = os.path.split(file)[0]


class FileDialog(object):

    @staticmethod
    def getSaveFileName(parent=None, title="Save File", file_dir=None, ext_filter="*"):
        if not file_dir:
            file_dir = _current_user_directory
        file_dlg = QtWidgets.QFileDialog.getSaveFileName(
            parent, title, file_dir, ext_filter
        )
        file = file_dlg[0] or None
        if file:
            _set_dir(file)
        return file_dlg

    @staticmethod
    def getOpenFileName(parent=None, title="Open File", file_dir=None, ext_filter="*"):
        if not file_dir:
            file_dir = _current_user_directory
        file_dlg = QtWidgets.QFileDialog.getOpenFileName(
            parent, title, file_dir, ext_filter
        )
        file = file_dlg[0] or None
        if file:
            _set_dir(file)
        return file_dlg


class BaseDialog(object):

    @staticmethod
    def message_dialog(
        parent=None, text="", title="Message", dialog_icon=None, custom_icon=None
    ):
        dlg = QtWidgets.QMessageBox(parent=parent)
        dlg.setWindowTitle(title)
        dlg.setInformativeText(text)
        dlg.setStandardButtons(QtWidgets.QMessageBox.Ok)

        if custom_icon:
            pixmap = QtGui.QPixmap(custom_icon).scaledToHeight(
                32, QtCore.Qt.SmoothTransformation
            )
            dlg.setIconPixmap(pixmap)
        else:
            if dialog_icon == "information":
                dlg.setIcon(dlg.Information)
            elif dialog_icon == "warning":
                dlg.setIcon(dlg.Warning)
            elif dialog_icon == "critical":
                dlg.setIcon(dlg.Critical)

        dlg.exec_()

    @staticmethod
    def question_dialog(
        parent=None, text="", title="Are you sure?", dialog_icon=None, custom_icon=None
    ):
        dlg = QtWidgets.QMessageBox(parent=parent)
        dlg.setWindowTitle(title)
        dlg.setInformativeText(text)
        dlg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if custom_icon:
            pixmap = QtGui.QPixmap(custom_icon).scaledToHeight(
                32, QtCore.Qt.SmoothTransformation
            )
            dlg.setIconPixmap(pixmap)
        else:
            if dialog_icon == "information":
                dlg.setIcon(dlg.Information)
            elif dialog_icon == "warning":
                dlg.setIcon(dlg.Warning)
            elif dialog_icon == "critical":
                dlg.setIcon(dlg.Critical)

        result = dlg.exec_()
        return bool(result == QtWidgets.QMessageBox.Yes)
